import signal
from logging import getLogger
from threading import Event
from typing import Any

logger = getLogger(__name__)


class SigHandler:
    _exit_event = Event()

    def __init__(self) -> None:
        """Registers SIGTERM, SIGINT handlers"""
        signal.signal(signal.SIGTERM, SigHandler._handler)
        signal.signal(signal.SIGINT, SigHandler._handler)

    @staticmethod
    def _handler(signum: int, stack: Any) -> None:
        logger.debug(f"Caught signal {signum}: setting exit event!")
        SigHandler._exit_event.set()

    @staticmethod
    def is_set() -> bool:
        """Return True if triggered; time to exit"""
        return SigHandler._exit_event.is_set()

    @staticmethod
    def wait_until_exit(timeout: float = 1.0) -> bool:
        """Sleep up to timeout seconds. Return True immediately if triggered."""
        return SigHandler._exit_event.wait(timeout=timeout)

    @staticmethod
    def set() -> None:
        """Trigger exit state; time to exit"""
        SigHandler._exit_event.set()



from .app_run import SubprocessAppRun


class OpenMPIRun(SubprocessAppRun):
    """
    https://www.open-mpi.org/doc/v3.0/man1/mpiexec.1.php
    """

    def _build_cmdline(self) -> str:
        node_ids = [h for h in self._node_spec.hostnames]
        env_args = [("-x", var) for var in self._envs.keys()]
        nid_str = ",".join(map(str, node_ids))
        args = [
            "mpiexec",
            "-n",
            self.get_num_ranks(),
            "--map-by",
            f"ppr:{self._ranks_per_node}:node",
            *[arg for pair in env_args for arg in pair],
            "-H",
            nid_str,
            self._cmdline,
        ]
        return " ".join(str(arg) for arg in args)



import concurrent.futures
import logging
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, NamedTuple, Optional, Set, Type, Union

from balsam import schemas
from balsam.schemas import JobState, deserialize, raise_from_serialized, serialize

from .app import ApplicationDefinition
from .manager import Manager
from .model import CreatableBalsamModel, Field, NonCreatableBalsamModel

if TYPE_CHECKING:
    from balsam._api.models import (  # noqa: F401
        App,
        BatchJob,
        EventLog,
        Job,
        JobManager,
        JobQuery,
        Session,
        Site,
        TransferItem,
    )
    from balsam.client import RESTClient

JobTransferItem = schemas.JobTransferItem
RUNNABLE_STATES = schemas.RUNNABLE_STATES
DONE_STATES = schemas.DONE_STATES

logger = logging.getLogger(__name__)


class SiteBase(CreatableBalsamModel):
    _create_model_cls = schemas.SiteCreate
    _update_model_cls = schemas.SiteUpdate
    _read_model_cls = schemas.SiteOut


class SiteManagerBase(Manager["Site"]):
    _api_path = "sites/"


class AppBase(CreatableBalsamModel):
    _create_model_cls = schemas.AppCreate
    _update_model_cls = schemas.AppUpdate
    _read_model_cls = schemas.AppOut


class AppManagerBase(Manager["App"]):
    _api_path = "apps/"


class BatchJobBase(CreatableBalsamModel):
    _create_model_cls = schemas.BatchJobCreate
    _update_model_cls = schemas.BatchJobUpdate
    _read_model_cls = schemas.BatchJobOut
    # These fields overwritten by the generated subclass anyway
    # Used for type checking the methods below
    queue: Field[str]
    project: Field[str]
    num_nodes: Field[int]
    wall_time_min: Field[int]
    partitions: Field[Optional[List[schemas.BatchJobPartition]]]
    optional_params: Field[Dict[str, str]]

    def validate(
        self,
        allowed_queues: Dict[str, schemas.AllowedQueue],
        allowed_projects: List[str],
        optional_batch_job_params: Dict[str, str],
    ) -> None:
        if self.queue not in allowed_queues:
            raise ValueError(f"Unknown queue {self.queue} " f"(known: {list(allowed_queues.keys())})")
        queue = allowed_queues[self.queue]
        if self.num_nodes > queue.max_nodes:
            raise ValueError(f"{self.num_nodes} exceeds queue max num_nodes {queue.max_nodes}")
        if self.num_nodes < 1:
            raise ValueError("self size must be at least 1 node")
        if self.wall_time_min > queue.max_walltime:
            raise ValueError(f"{self.wall_time_min} exceeds queue max wall_time_min {queue.max_walltime}")

        if self.project not in allowed_projects:
            raise ValueError(f"Unknown project {self.project} " f"(known: {allowed_projects})")
        if self.partitions:
            if sum(part.num_nodes for part in self.partitions) != self.num_nodes:
                raise ValueError("Sum of partition sizes must equal batchjob num_nodes")

        extras = set(self.optional_params.keys())
        allowed_extras = set(optional_batch_job_params.keys())
        extraneous = extras.difference(allowed_extras)
        if extraneous:
            raise ValueError(f"Extraneous optional_params: {extraneous} " f"(allowed extras: {allowed_extras})")

    def partitions_to_cli_args(self) -> str:
        if not self.partitions:
            return ""
        args = ""
        for part in self.partitions:
            job_mode = part.job_mode
            num_nodes = part.num_nodes
            filter_tags = ":".join(f"{k}={v}" for k, v in part.filter_tags.items())
            args += f" --part {job_mode}:{num_nodes}"
            if filter_tags:
                args += f":{filter_tags}"
        return args


class BatchJobManagerBase(Manager["BatchJob"]):
    _api_path = "batch-jobs/"
    _bulk_update_enabled = True


AppDefType = Type[ApplicationDefinition]
InputAppType = Union[int, str, AppDefType]


class JobBase(CreatableBalsamModel):
    _create_model: Optional[schemas.JobCreate]
    _update_model: Optional[schemas.JobUpdate]
    _read_model: Optional[schemas.JobOut]
    _create_model_cls = schemas.JobCreate
    _update_model_cls = schemas.JobUpdate
    _read_model_cls = schemas.JobOut

    objects: "JobManager"
    app_id: Field[int]
    workdir: Field[Path]
    parent_ids: Field[Set[int]]
    state: Field[Optional[JobState]]

    class NoResult(ValueError):
        pass

    def __init__(self, app_id: InputAppType, site_name: Optional[str] = None, **kwargs: Any) -> None:
        app_id = self._resolve_app_id(app_id, site_name)
        super().__init__(**kwargs, app_id=app_id)

    @classmethod
    def _resolve_app_id(cls, app: InputAppType, site_name: Optional[str]) -> int:
        if isinstance(app, int):
            app_id = app
        elif isinstance(app, str):
            app = ApplicationDefinition.load_by_name(app, site_name)
            assert app.__app_id__ is not None
            app_id = app.__app_id__
        else:
            if app.__app_id__ is None:
                raise ValueError(
                    f"Cannot resolve ID from ApplicationDefinition {app}: __app_id__ is None. You need to app.sync() prior to creating Jobs with this app."
                )
            app_id = app.__app_id__
        return app_id

    @property
    def app(self) -> AppDefType:
        """
        Fetch the ApplicationDefinition class associated with this Job.
        """
        if self.app_id is None:
            raise ValueError("Cannot fetch by app ID; is None")
        return ApplicationDefinition.load_by_id(self.app_id)

    @property
    def site_id(self) -> int:
        """
        Fetch the Site ID associated with Job
        """
        if self.app_id is None:
            raise ValueError("Cannot fetch by app ID; is None")
        app_def = ApplicationDefinition.load_by_id(self.app_id)
        assert app_def._site_id is not None
        return app_def._site_id

    def get_parameters(self) -> Dict[str, Any]:
        """
        Unpack and return the Job parameters dictionary
        """
        if self._state == "clean":
            assert self._read_model is not None
            ser = self._read_model.serialized_parameters
        elif self._state == "creating":
            assert self._create_model is not None
            ser = self._create_model.serialized_parameters
        else:
            if "serialized_parameters" in self._dirty_fields:
                assert self._update_model is not None
                ser = self._update_model.serialized_parameters
            else:
                assert self._read_model is not None
                ser = self._read_model.serialized_parameters
        params: Dict[str, Any] = deserialize(ser)
        if not isinstance(params, dict):
            raise ValueError(f"Deserialized Job parameters are of type {type(params)}; must be dict.")
        return params

    def set_parameters(self, value: Dict[str, Any]) -> None:
        """
        Set the Job parameters dictionary
        """
        serialized = serialize(value)
        if self._state == "creating":
            assert self._create_model is not None
            self._create_model.serialized_parameters = serialized
        else:
            if self._update_model is None:
                self._update_model = self._update_model_cls()
            self._update_model.serialized_parameters = serialized
            self._dirty_fields.add("serialized_parameters")
            self._state = "dirty"

    def result_nowait(self) -> Any:
        """
        Unpack and return the Job result, or re-raise the exception.
        Raises `Job.NoResult` if there is no return value or exception yet.
        """
        if self._read_model is None:
            raise ValueError("Job data not yet loaded from API")
        s_ret = self._read_model.serialized_return_value
        s_exc = self._read_model.serialized_exception
        if s_ret:
            return deserialize(s_ret)
        elif s_exc:
            raise_from_serialized(s_exc)
        else:
            raise JobBase.NoResult("No return value or exception has been reported")

    def result(self, timeout: Optional[float] = None) -> Any:
        """
        Unpack and return the Job result, or re-raise the exception.
        Blocks for `timeout` sec and raises `concurrent.futures.TimeoutError` if there is no return value or exception yet.
        """
        wait_for: List["Job"] = [self]  # type: ignore
        res = self.objects.wait(wait_for, timeout=timeout)
        if res.done:
            return self.result_nowait()
        else:
            raise concurrent.futures.TimeoutError(f"Job is still {self.state} after {timeout} sec timeout")

    def done(self) -> bool:
        """
        Refresh the Job and return True if it has completed processing (either "JOB_FINISHED" or "FAILED" state.)
        """
        self.refresh_from_db()
        return self.state in [JobState.job_finished, JobState.failed]

    def resolve_workdir(self, data_path: Path) -> Path:
        return data_path.joinpath(self.workdir)

    def parent_query(self) -> "JobQuery":
        """
        Returns a JobQuery for the parents of this Job.
        """
        return self.objects.filter(id=list(self.parent_ids))


class JobWaitResult(NamedTuple):
    done: List["Job"]
    not_done: List["Job"]


class JobManagerBase(Manager["Job"]):
    _api_path = "jobs/"
    _bulk_create_enabled = True
    _bulk_update_enabled = True
    _bulk_delete_enabled = True

    def bulk_refresh(self, jobs: List["Job"]) -> None:
        """
        Refresh the list of Jobs from the latest database state
        """
        job_manager: "JobManager" = self  # type: ignore
        jobs_by_id = {job.id: job for job in jobs if job.id is not None}
        fetched = job_manager.filter(id=list(jobs_by_id.keys()))
        for api_job in fetched:
            assert api_job.id is not None and api_job._read_model is not None
            jobs_by_id[api_job.id]._refresh_from_dict(api_job._read_model.dict())

    def wait(
        self,
        jobs: List["Job"],
        timeout: Optional[float] = None,
        poll_interval: float = 1.0,
        return_when: str = "ALL_COMPLETED",
    ) -> JobWaitResult:
        """
        Block and periodically refresh jobs, until either all have completed
        (the default) or any have completed (return_when="FIRST_COMPLETED").
        Also returns after `timeout` seconds.  Rather than raising a Timeout
        error, returns a named tuple containing `done` and `not_done` Job lists.
        """

        start = time.time()
        done_jobs: List["Job"] = []
        not_done_jobs: List["Job"] = []

        for job in jobs:
            if job.state in DONE_STATES:
                done_jobs.append(job)
            else:
                not_done_jobs.append(job)

        def should_exit() -> bool:
            timed_out = (time.time() - start > timeout) if timeout is not None else False
            if return_when == "ALL_COMPLETED":
                return timed_out or len(not_done_jobs) == 0
            else:
                return timed_out or len(done_jobs) > 0

        while not should_exit():
            time.sleep(poll_interval)
            self.bulk_refresh(not_done_jobs)
            finished = [job for job in not_done_jobs if job.state in DONE_STATES]
            not_done_jobs[:] = [job for job in not_done_jobs if job.state not in DONE_STATES]
            done_jobs.extend(finished)

        return JobWaitResult(done=done_jobs, not_done=not_done_jobs)

    def as_completed(
        self,
        jobs: List["Job"],
        timeout: Optional[float] = None,
        poll_interval: float = 1.0,
    ) -> Iterator["Job"]:
        """
        Returns an iterator over the Job instances as they complete.  Raises a
        `concurrent.futures.TimeoutError` if __next__() is called and the result
        isn’t available after timeout seconds from the original call to
        as_completed().
        """

        start = time.time()
        pending_jobs: List["Job"] = []

        for job in jobs:
            if job.state in DONE_STATES:
                yield job
            else:
                pending_jobs.append(job)

        def should_exit() -> bool:
            timed_out = (time.time() - start > timeout) if timeout is not None else False
            return timed_out or not pending_jobs

        while not should_exit():
            time.sleep(poll_interval)
            self.bulk_refresh(pending_jobs)
            for job in pending_jobs:
                if job.state in DONE_STATES:
                    yield job
            pending_jobs[:] = [job for job in pending_jobs if job.state not in DONE_STATES]

        if pending_jobs:
            raise concurrent.futures.TimeoutError(
                f"{len(pending_jobs)} Jobs are still pending after {timeout} sec timeout."
            )


class SessionBase(CreatableBalsamModel):
    _create_model_cls = schemas.SessionCreate
    _update_model_cls = None
    _read_model_cls = schemas.SessionOut
    objects: "SessionManagerBase"

    def acquire_jobs(
        self,
        max_num_jobs: int,
        max_wall_time_min: Optional[int] = None,
        max_nodes_per_job: Optional[int] = None,
        max_aggregate_nodes: Optional[float] = None,
        serial_only: bool = False,
        filter_tags: Optional[Dict[str, str]] = None,
        states: Set[JobState] = RUNNABLE_STATES,
        app_ids: Optional[Set[int]] = None,
    ) -> "List[Job]":
        if filter_tags is None:
            filter_tags = {}
        if app_ids is None:
            app_ids = set()
        return self.__class__.objects._do_acquire(
            self,
            max_num_jobs=max_num_jobs,
            max_wall_time_min=max_wall_time_min,
            max_nodes_per_job=max_nodes_per_job,
            max_aggregate_nodes=max_aggregate_nodes,
            serial_only=serial_only,
            filter_tags=filter_tags,
            states=states,
            app_ids=app_ids,
        )

    def tick(self) -> None:
        return self.__class__.objects._do_tick(self)


class SessionManagerBase(Manager["Session"]):
    _api_path = "sessions/"
    _client: "RESTClient"

    def _do_acquire(self, instance: "SessionBase", **kwargs: Any) -> "List[Job]":
        from balsam._api.models import Job, JobManager

        Job.objects = JobManager(client=self._client)

        acquired_raw = self._client.post(self._api_path + f"{instance.id}", **kwargs)
        jobs = [Job._from_api(dat) for dat in acquired_raw]
        return jobs

    def _do_tick(self, instance: "SessionBase") -> None:
        self._client.put(self._api_path + f"{instance.id}")


class TransferItemBase(NonCreatableBalsamModel):
    _create_model_cls = None
    _update_model_cls = schemas.TransferItemUpdate
    _read_model_cls = schemas.TransferItemOut


class TransferItemManagerBase(Manager["TransferItem"]):
    _api_path = "transfers/"
    _bulk_update_enabled = True


class EventLogBase(NonCreatableBalsamModel):
    _create_model_cls = None
    _update_model_cls = None
    _read_model_cls = schemas.LogEventOut


class EventLogManagerBase(Manager["EventLog"]):
    _api_path = "events/"



from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional
from uuid import UUID

from pydantic import AnyUrl, BaseModel, Field, validator

from .batchjob import SchedulerBackfillWindow, SchedulerJobStatus


class QueuedJobState(str, Enum):
    queued = "queued"
    running = "running"


class TransferProtocol(str, Enum):
    globus = "globus"
    rsync = "rsync"


class BackfillWindow(BaseModel):
    queue: str = Field(..., example="default", description="Queue name to which Jobs may be submitted")
    num_nodes: int = Field(..., example=130, description="Number of idle nodes")
    wall_time_min: int = Field(..., example=40, description="Duration in minutes that nodes will remain idle")


class QueuedJob(BaseModel):
    queue: str = Field(..., example="default", description="Name of queue")
    num_nodes: int = Field(..., example=128, description="Number of nodes allocated")
    wall_time_min: int = Field(..., example=60, description="Length of node allocation")
    state: QueuedJobState = Field(..., example=QueuedJobState.queued, description="Status of the allocation")


class AllowedQueue(BaseModel):
    max_nodes: int = Field(..., example=8)
    max_walltime: int = Field(..., example=60)
    max_queued_jobs: int = Field(..., example=1)


class SiteBase(BaseModel):
    name: str = Field(..., example="theta-knl", description="Unique Site name")
    path: Path = Field(
        ..., example="/projects/datascience/user/mySite", description="Absolute filesystem path of the Site"
    )
    globus_endpoint_id: Optional[UUID] = Field(None, description="Associated Globus endpoint ID")
    backfill_windows: Dict[str, List[SchedulerBackfillWindow]] = Field(
        {}, description="Idle backfill currently available at the Site, keyed by queue name"
    )
    queued_jobs: Dict[int, SchedulerJobStatus] = Field(
        {}, description="Queued scheduler jobs at the Site, keyed by scheduler ID"
    )
    optional_batch_job_params: Dict[str, str] = Field(
        {},
        example={"enable_ssh": 1},
        description="Optional pass-through parameters accepted by the Site batchjob template",
    )
    allowed_projects: List[str] = Field(
        [],
        example=["datascience", "materials-adsp"],
        description="Allowed projects/allocations for batchjob submission",
    )
    allowed_queues: Dict[str, AllowedQueue] = Field(
        {},
        example={
            "debug-cache-quad": {
                "max_nodes": 8,
                "max_walltime": 60,
                "max_queued_jobs": 1,
            }
        },
        description="Allowed queues and associated queueing policies",
    )
    transfer_locations: Dict[str, AnyUrl] = Field(
        {},
        example={
            "APS-DTN": "globus://ddb59aef-6d04-11e5-ba46-22000b92c6ec",
            "MyCluster": "rsync://user@hostname.mycluster",
        },
        description="Trusted transfer location aliases and associated protocol/URLs",
    )

    @validator("path")
    def path_is_absolute(cls, v: Path) -> Path:
        if not v.is_absolute():
            raise ValueError("path must be absolute")
        return v


class SiteCreate(SiteBase):
    pass


class SiteUpdate(SiteBase):
    name: Optional[str] = Field(None, example="theta-knl", description="Unique Site name")  # type: ignore
    path: Optional[Path] = Field(None, example="/projects/datascience/user/mySite", description="Absolute filesystem path of the Site")  # type: ignore


class SiteOut(SiteBase):
    class Config:
        orm_mode = True

    id: int = Field(..., example=123)
    last_refresh: datetime = Field(..., example=datetime.utcnow())
    creation_date: datetime = Field(..., example=datetime.utcnow())


class PaginatedSitesOut(BaseModel):
    count: int
    results: List[SiteOut]



from datetime import datetime
from typing import Any, Dict, List

from fastapi import APIRouter, Depends, status
from sqlalchemy import orm

from balsam import schemas
from balsam.server import ValidationError, settings
from balsam.server.models import Job, crud, get_session
from balsam.server.pubsub import pubsub
from balsam.server.utils import Paginator

from .filters import JobQuery

router = APIRouter()
auth = settings.auth.get_auth_method()


@router.get("/", response_model=schemas.PaginatedJobsOut)
def list(
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
    paginator: Paginator[Job] = Depends(Paginator),
    q: JobQuery = Depends(JobQuery),
) -> Dict[str, Any]:
    """List the user's Jobs."""
    count, jobs = crud.jobs.fetch(db, owner=user, paginator=paginator, filterset=q)
    return {"count": count, "results": jobs}


@router.get("/{job_id}", response_model=schemas.JobOut)
def read(job_id: int, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)) -> Job:
    """Get a Job by id."""
    count, jobs = crud.jobs.fetch(db, owner=user, job_id=job_id)
    return jobs[0]


@router.post("/", response_model=List[schemas.JobOut], status_code=status.HTTP_201_CREATED)
def bulk_create(
    jobs: List[schemas.ServerJobCreate], db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)
) -> List[schemas.JobOut]:
    """Create a list of Jobs."""
    new_jobs, new_events, new_transfers = crud.jobs.bulk_create(db, owner=user, job_specs=jobs)

    result_jobs = [schemas.JobOut.from_orm(job) for job in new_jobs]
    result_events = [schemas.LogEventOut.from_orm(e) for e in new_events]
    result_transfers = [schemas.TransferItemOut.from_orm(t) for t in new_transfers]

    db.commit()
    pubsub.publish(user.id, "bulk-create", "job", result_jobs)
    pubsub.publish(user.id, "bulk-create", "event", result_events)
    pubsub.publish(user.id, "bulk-create", "transfer-item", result_transfers)
    return result_jobs


@router.patch("/", response_model=List[schemas.JobOut])
def bulk_update(
    jobs: List[schemas.JobBulkUpdate],
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
) -> List[schemas.JobOut]:
    """Update a list of Jobs"""
    now = datetime.utcnow()
    patch_dicts = {job.id: {**job.dict(exclude_unset=True, exclude={"id"}), "last_update": now} for job in jobs}
    if len(jobs) > len(patch_dicts):
        raise ValidationError("Duplicate Job ID keys provided")
    updated_jobs, new_events = crud.jobs.bulk_update(db, owner=user, patch_dicts=patch_dicts)

    result_jobs = [schemas.JobOut.from_orm(job) for job in updated_jobs]
    result_events = [schemas.LogEventOut.from_orm(e) for e in new_events]
    db.commit()

    pubsub.publish(user.id, "bulk-update", "job", result_jobs)
    pubsub.publish(user.id, "bulk-create", "event", result_events)
    return result_jobs


@router.put("/", response_model=List[schemas.JobOut])
def query_update(
    update_fields: schemas.JobUpdate,
    db: orm.Session = Depends(get_session),
    user: schemas.UserOut = Depends(auth),
    q: JobQuery = Depends(JobQuery),
) -> List[schemas.JobOut]:
    """Apply the same update to all Jobs selected by the query."""
    data = update_fields.dict(exclude_unset=True)
    data["last_update"] = datetime.utcnow()
    updated_jobs, new_events = crud.jobs.update_query(db, owner=user, update_data=data, filterset=q)

    result_jobs = [schemas.JobOut.from_orm(job) for job in updated_jobs]
    result_events = [schemas.LogEventOut.from_orm(e) for e in new_events]
    db.commit()

    pubsub.publish(user.id, "bulk-update", "job", result_jobs)
    pubsub.publish(user.id, "bulk-create", "event", result_events)
    return result_jobs


@router.put("/{job_id}", response_model=schemas.JobOut)
def update(
    job_id: int, job: schemas.JobUpdate, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)
) -> schemas.JobOut:
    """Update a Job by id."""
    data = job.dict(exclude_unset=True)
    data["last_update"] = datetime.utcnow()
    patch = {job_id: data}
    updated_jobs, new_events = crud.jobs.bulk_update(db, owner=user, patch_dicts=patch)

    result_jobs = [schemas.JobOut.from_orm(job) for job in updated_jobs]
    result_events = [schemas.LogEventOut.from_orm(e) for e in new_events]

    db.commit()
    pubsub.publish(user.id, "bulk-update", "job", result_jobs)
    pubsub.publish(user.id, "bulk-create", "event", result_events)
    return result_jobs[0]


@router.delete("/{job_id}", status_code=status.HTTP_204_NO_CONTENT)
def delete(job_id: int, db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth)) -> None:
    """Delete a Job by id."""
    crud.jobs.delete_query(db, owner=user, job_id=job_id)
    db.commit()
    pubsub.publish(user.id, "bulk-delete", "job", {"ids": [job_id]})


@router.delete("/", status_code=status.HTTP_204_NO_CONTENT)
def query_delete(
    db: orm.Session = Depends(get_session), user: schemas.UserOut = Depends(auth), q: JobQuery = Depends(JobQuery)
) -> None:
    """Delete all jobs selected by the query."""
    crud.jobs.delete_query(db, owner=user, filterset=q)
    db.commit()



from datetime import timedelta
from typing import Any, Dict, List, Optional, Type

from balsam._api.models import (
    App,
    AppManager,
    BatchJob,
    BatchJobManager,
    EventLog,
    EventLogManager,
    Job,
    JobManager,
    Session,
    SessionManager,
    Site,
    SiteManager,
    TransferItem,
    TransferItemManager,
)

from .encoders import jsonable_encoder


class AuthError(Exception):
    pass


class RESTClient:
    expires_in: timedelta

    def __init__(*args: Any, **kwargs: Any) -> None:
        raise NotImplementedError

    def interactive_login(self) -> Any:
        """Initiate interactive login flow"""
        raise NotImplementedError

    def refresh_auth(self) -> None:
        """
        Reload credentials if stored/not expired.
        Set appropriate Auth headers on HTTP session.
        """
        raise NotImplementedError

    def close_session(self) -> None:
        raise NotImplementedError

    def request(
        self,
        url: str,
        http_method: str,
        params: Optional[Dict[str, Any]] = None,
        json: Any = None,
        data: Any = None,
        authenticating: bool = False,
    ) -> Any:
        """
        Supports timeout retry, auto re-authentication, accepting DUPLICATE status
        Raises helfpul errors on 4**, 5**, TimeoutErrors, AuthErrors
        """
        raise NotImplementedError

    def get(self, url: str, **kwargs: Any) -> Any:
        """GET kwargs become URL query parameters (e.g. /?site=3)"""
        return self.request(url, "GET", params=kwargs)

    def post_form(self, url: str, **kwargs: Any) -> Any:
        return self.request(url, "POST", data=kwargs)

    def post(self, url: str, authenticating: bool = False, **kwargs: Any) -> Any:
        return self.request(url, "POST", json=jsonable_encoder(kwargs), authenticating=authenticating)

    def bulk_post(self, url: str, list_data: List[Any]) -> Any:
        return self.request(url, "POST", json=jsonable_encoder(list_data))

    def put(self, url: str, **kwargs: Any) -> Any:
        return self.request(url, "PUT", json=jsonable_encoder(kwargs))

    def bulk_put(self, url: str, payload: Any, **kwargs: Any) -> Any:
        return self.request(url, "PUT", json=jsonable_encoder(payload), params=kwargs)

    def patch(self, url: str, **kwargs: Any) -> Any:
        return self.request(url, "PATCH", json=jsonable_encoder(kwargs))

    def bulk_patch(self, url: str, list_data: List[Any]) -> Any:
        return self.request(url, "PATCH", json=jsonable_encoder(list_data))

    def delete(self, url: str) -> Any:
        return self.request(url, "DELETE")

    def bulk_delete(self, url: str, **kwargs: Any) -> Any:
        return self.request(url, "DELETE", params=kwargs)

    @property
    def Site(self) -> Type[Site]:
        return SiteManager(client=self)._model_class

    @property
    def App(self) -> Type[App]:
        return AppManager(client=self)._model_class

    @property
    def BatchJob(self) -> Type[BatchJob]:
        return BatchJobManager(client=self)._model_class

    @property
    def Job(self) -> Type[Job]:
        return JobManager(client=self)._model_class

    @property
    def TransferItem(self) -> Type[TransferItem]:
        return TransferItemManager(client=self)._model_class

    @property
    def Session(self) -> Type[Session]:
        return SessionManager(client=self)._model_class

    @property
    def EventLog(self) -> Type[EventLog]:
        return EventLogManager(client=self)._model_class



import copy
import logging
import os
import random
import time
from pathlib import Path
from typing import Any
from uuid import uuid4

from locust import User, task  # type: ignore
from locust.env import Environment  # type: ignore
from locust.event import EventHook  # type: ignore

from balsam._api import models
from balsam.client import OAuthRequestsClient

logger = logging.getLogger(__name__)

TEST_TOKENS_FILE = os.environ.get("BALSAM_TEST_TOKENS", "test-tokens.txt")
if not Path(TEST_TOKENS_FILE).is_file():
    raise RuntimeError(
        f"Cannot find token file {TEST_TOKENS_FILE}. Create this or set the path with BALSAM_TEST_TOKENS."
    )

with open(TEST_TOKENS_FILE) as fp:
    tokens = [line.strip() for line in fp.readlines()]


class LocustBalsamClient(OAuthRequestsClient):
    def __init__(self, api_root: str, request_event: EventHook) -> None:
        super().__init__(api_root, token=random.choice(tokens))
        self._request_event = request_event

    def request(self, url: str, http_method: str, **kwargs: Any) -> Any:  # type: ignore
        start_time = time.perf_counter()
        url_parts = [(part if not part.isdigit() else ":id") for part in url.split("/")]
        url_name = "/".join(url_parts)
        request_meta = {
            "request_type": "balsam-requests-client",
            "name": f"{http_method} {url_name}",
            "response_length": 0,
            "response": None,
            "context": {},  # see HttpUser if you actually want to implement contexts
            "exception": None,
        }
        try:
            request_meta["response"] = super().request(url, http_method, **kwargs)
        except Exception as e:
            request_meta["exception"] = e
            raise
        finally:
            request_meta["response_time"] = (time.perf_counter() - start_time) * 1000
            self._request_event.fire(**request_meta)  # This is what makes the request actually get logged in Locust
        return request_meta["response"]


class BalsamUser(User):  # type: ignore
    def __init__(self, environment: Environment) -> None:
        super().__init__(environment)
        self.client = LocustBalsamClient("https://balsam-dev.alcf.anl.gov/", environment.events.request)
        self.Site = copy.deepcopy(models.Site)
        self.Site.objects = models.SiteManager(self.client)
        self.App = copy.deepcopy(models.App)  # Deepcopy
        self.App.objects = models.AppManager(self.client)
        self.BatchJob = copy.deepcopy(models.BatchJob)  # Deepcopy
        self.BatchJob.objects = models.BatchJobManager(self.client)
        self.Job = copy.deepcopy(models.Job)  # Deepcopy
        self.Job.objects = models.JobManager(self.client)
        self.TransferItem = copy.deepcopy(models.TransferItem)  # Deepcopy
        self.TransferItem.objects = models.TransferItemManager(self.client)
        self.Session = copy.deepcopy(models.Session)  # Deepcopy
        self.Session.objects = models.SessionManager(self.client)
        self.EventLog = copy.deepcopy(models.EventLog)  # Deepcopy
        self.EventLog.objects = models.EventLogManager(self.client)

    def on_start(self) -> None:
        # Clean up old Sites prior to testing
        for site in self.Site.objects.all():
            logger.debug(f"Deleting old site: {site}")
            site.delete()
            assert site.id is None
        logger.debug("OK I should have deleted all my sites by now!")
        failed_to_delete = list(self.Site.objects.all())
        assert len(failed_to_delete) == 0
        name = f"test-site{uuid4()}"
        path = Path("/projects/premade_site")
        self.premade_site = self.Site.objects.create(hostname=name, path=path)
        assert self.premade_site.id is not None
        class_path = "premade_app"
        parameters = {"foo": {"required": False, "default": "foo"}, "bar": {"required": False, "default": "bar"}}
        self.premade_app = self.App.objects.create(
            site_id=self.premade_site.id, class_path=class_path, parameters=parameters  # type: ignore
        )

    def on_stop(self) -> None:
        # Clean up Sites after testing
        for site in self.Site.objects.all():
            site.delete()

    @task(1)
    def site_create(self) -> None:
        num_sites = len(self.Site.objects.all())
        name = f"site_{num_sites}"
        path = f"/projects/foo/{uuid4()}"
        self.Site.objects.create(hostname=name, path=path)

    @task(3)
    def site_list(self) -> None:
        list(self.Site.objects.all())

    @task(1)
    def app_create(self):
        site = self.premade_site
        num_apps = len(self.App.objects.filter(site_id=site.id))
        class_path = f"foo.bar_{num_apps:08d}"
        parameters = {"foo": {"required": False, "default": "foo"}, "bar": {"required": False, "default": "bar"}}
        self.App.objects.create(
            site_id=site.id,
            class_path=class_path,
            parameters=parameters,
        )

    @task(3)
    def app_list(self) -> None:
        list(self.App.objects.all())

    @task(10)
    def job_create(self) -> None:
        app = self.premade_app

        parameters = {"foo": "yes", "bar": "maybe"}
        job_num = len(self.Job.objects.all())
        self.Job.objects.create("test/run_%08d" % job_num, app_id=app.id, parameters=parameters)

    @task(10)
    def job_list(self) -> None:
        list(self.Job.objects.all())

    @task(10)
    def job_list_site(self) -> None:
        list(self.Job.objects.filter(site_id=self.premade_site.id))

    @task(10)
    def bulk_job_submission(self) -> None:
        site = self.premade_site
        app = self.premade_app
        # create a bunch of jobs:
        parameters = {"foo": "yes", "bar": "maybe"}
        batch_id = str(int(time.time()))
        # create between 1 and 100 jobs
        jobs = [
            self.Job(
                f"batchjob_{batch_id}/run_{i:08d}",
                app_id=app.id,
                parameters=parameters,
                tags={"batch_id": batch_id},
            )
            for i in range(random.choice(range(128, 512)))
        ]
        jobs = self.Job.objects.bulk_create(jobs)

        # move jobs to PREPROCESSED state
        for job in jobs:
            job.state = "PREPROCESSED"
        self.Job.objects.bulk_update(jobs)

        # create batch job
        simulated_nodes = 128
        batch_job = self.BatchJob.objects.create(
            site_id=site.id,
            project="datascience",
            queue="default",
            num_nodes=simulated_nodes,
            wall_time_min=30,
            job_mode="mpi",
        )
        # create session
        sess = self.Session.objects.create(batch_job_id=batch_job.id, site_id=site.id)

        # associate jobs with batchjob
        sess.acquire_jobs(
            max_num_jobs=1024,
            filter_tags={"batch_id": batch_id},
            states=["PREPROCESSED", "RESTART_READY"],
        )

        # session heartbeat
        sess.tick()

        steps = int(simulated_nodes / len(jobs)) + 1
        for step in range(steps):

            start = simulated_nodes * step
            end = min(simulated_nodes * (step + 1), len(jobs))

            # change jobs to running state
            for job in jobs[start:end]:
                job.state = "RUNNING"
            self.Job.objects.bulk_update(jobs[start:end])

            # session heartbeat
            sess.tick()

            # change jobs to done or error state
            for job in jobs[start:end]:
                job.state = "RUN_DONE" if random.random() < 0.95 else "RUN_ERROR"
            self.Job.objects.bulk_update(jobs[start:end])

            # session heartbeat
            sess.tick()

        # Delete session when exiting:
        sess.delete()



"""
Clients: perform requests to Balsam server
"""
from .requests_client import NotAuthenticatedError, RequestsClient
from .requests_oauth import OAuthRequestsClient
from .requests_password import BasicAuthRequestsClient
from .rest_base_client import RESTClient

__all__ = [
    "RESTClient",
    "RequestsClient",
    "BasicAuthRequestsClient",
    "OAuthRequestsClient",
    "NotAuthenticatedError",
]



from datetime import datetime

import pytest

from balsam._api.models import Job
from balsam.api import ApplicationDefinition
from balsam.platform.app_run import MPICHRun
from balsam.platform.compute_node import ThetaGPUNode
from balsam.site.launcher._mpi_mode import Launcher
from balsam.site.launcher.node_manager import NodeManager


class HelloWorld(ApplicationDefinition):
    site = 0
    command_template = "echo Hello, {{name}}"
    parameters = {}


@pytest.fixture(scope="function")
def launcher(mocker, tmp_path):
    mocker.patch("balsam.platform.app_run.app_run.SubprocessAppRun", autospec=True)
    mock_job_source = mocker.patch("balsam.site.job_source.SynchronousJobSource", autospec=True)
    mock_status_updater = mocker.patch("balsam.site.status_updater.BulkStatusUpdater", autospec=True)
    mock_client = mocker.patch("balsam.client.BasicAuthRequestsClient", autospec=True)
    client = mock_client("http://test:1234", "testuser", token="foo")

    nodes = [ThetaGPUNode(nid, "thetagpu{nid:02d}") for nid in range(1, 12)]
    node_manager = NodeManager(nodes, allow_node_packing=True)

    mock_job_source.get_jobs.return_value = [
        Job(
            _api_data=True,
            id=i,
            workdir=f"test/{i}",
            app_id=1,
            state="PREPROCESSED",
            serialized_parameters="",
            serialized_exception="",
            serialized_return_value="",
            last_update=datetime.utcnow(),
            pending_file_cleanup=True,
        )
        for i in range(10)
    ]

    job_source = mock_job_source(
        client=client,
        site_id=123,
        filter_tags={},
        max_wall_time_min=60,
        scheduler_id=25,
    )
    status_updater = mock_status_updater(client)
    return Launcher(
        data_dir=tmp_path,
        idle_ttl_sec=10,
        delay_sec=0.01,
        app_run=MPICHRun,
        node_manager=node_manager,
        job_source=job_source,
        status_updater=status_updater,
        wall_time_min=60,
        error_tail_num_lines=10,
        max_concurrent_runs=1000,
    )


def test_launcher_foo(launcher, tmp_path):
    launcher.launch_runs()



import click
import yaml

from balsam.config import ClientSettings

from .utils import filter_by_sites


@click.group()
def app() -> None:
    """
    Manage Balsam applications
    """
    pass


@app.command()
@click.option("-v", "--verbose", is_flag=True)
@click.option("-s", "--site", "site_selector", default="")
def ls(site_selector: str, verbose: bool) -> None:
    """
    List Apps

    1) View apps across all sites

        balsam app ls --site=all

    2) Filter apps by specific site IDs or Path fragments

        balsam app ls --site=123,my_site_folder
    """
    client = ClientSettings.load_from_file().build_client()
    qs = client.App.objects.all()
    qs = filter_by_sites(qs, site_selector)
    if verbose:
        reprs = [yaml.dump(app.display_dict(), sort_keys=False, indent=4) for app in qs]
        print(*reprs, sep="\n----\n")
    else:
        sites = {site.id: site for site in client.Site.objects.all()}
        click.echo(f"{'ID':>5s}   {'Name':>18s}   {'Site':<20s}")
        apps = sorted(list(qs), key=lambda app: app.site_id)
        for a in apps:
            site = sites[a.site_id]
            click.echo(f"{a.id:>5d}   {a.name:>18s}   {site.name:<20s}")


@app.command()
@click.option("-n", "--name", required=True)
@click.option("-s", "--site", "site_selector", default="")
def rm(site_selector: str, name: str) -> None:
    client = ClientSettings.load_from_file().build_client()
    qs = client.App.objects.all()
    qs = filter_by_sites(qs, site_selector)

    resolved_app = qs.get(name=name)
    resolved_id = resolved_app.id
    appstr = f"App(id={resolved_id}, name={resolved_app.name})"
    job_count = client.Job.objects.filter(app_id=resolved_id).count()
    if job_count == 0:
        resolved_app.delete()
        click.echo(f"Deleted {appstr}: there were no associated jobs.")
    elif click.confirm(f"Really Delete {appstr}?? There are {job_count} Jobs that will be ERASED!"):
        resolved_app.delete()
        click.echo(f"Deleted App {resolved_id} ({name})")



import shutil
import socket
import sys
from pathlib import Path
from typing import Optional, Union

import click

from balsam.config import ClientSettings, InvalidSettings, Settings, SiteConfig, site_builder
from balsam.util import globus_auth

from .utils import (
    PID_FILENAME,
    check_killable,
    get_pidfile,
    is_site_active,
    kill_site,
    load_site_config,
    read_pidfile,
    start_site,
)


@click.group()
def site() -> None:
    """
    Setup or manage your Balsam sites
    """
    pass


@site.command()
def start() -> None:
    """Start up the site daemon"""
    cf = load_site_config()
    if not get_pidfile(cf).is_file():
        click.echo(f"Starting Balsam site at {cf.site_path}")
        proc = start_site(cf.site_path)
        click.echo(f"Started Balsam site daemon [pid {proc.pid}] on {socket.gethostname()}")
        return None

    host: Optional[str]
    pid: Optional[int]
    try:
        host, pid = read_pidfile(cf)
    except Exception:
        raise click.BadArgumentUsage(
            f"{PID_FILENAME} already exists in {cf.site_path}: "
            "This means the site is already running; to restart it, "
            "first use `balsam site stop`."
        )
    else:
        raise click.BadArgumentUsage(
            f"The site is already running on {host} [PID {pid}].\nTo restart it, first use `balsam site stop`."
        )


@site.command()
def stop() -> None:
    """Stop site daemon"""
    cf = load_site_config()
    kill_pid = check_killable(cf, raise_exc=True)
    if kill_pid is not None:
        kill_site(cf, kill_pid)


@site.command()
@click.argument("site-path", type=click.Path(writable=True))
@click.option("-n", "--name", prompt="Enter Site Name", default=socket.gethostname, show_default=socket.gethostname())
def init(site_path: Union[str, Path], name: str) -> None:
    """
    Create a new balsam site at SITE-PATH

    balsam site init path/to/site
    """
    import inquirer  # type: ignore

    site_path = Path(site_path).absolute()
    if site_path.exists():
        raise click.BadParameter(f"{site_path} already exists")

    path_conf_map = site_builder.load_default_configs()
    site_prompt = inquirer.List(
        "path",
        message="Select system",
        choices=[(conf.title, path) for path, conf in path_conf_map.items()],
        carousel=True,
    )
    selected_path = inquirer.prompt([site_prompt])["path"]

    try:
        site_builder.new_site_setup(
            site_path=site_path,
            default_site_path=selected_path,
            default_site_conf=path_conf_map[selected_path],
            name=name,
        )
    except (InvalidSettings, FileNotFoundError) as exc:
        click.echo(str(exc))
        sys.exit(1)
    click.echo(f"New Balsam site set up at {site_path}")


@site.command()
@click.argument("src", type=click.Path(exists=True, file_okay=False))
@click.argument("dest", type=click.Path(exists=False, writable=True))
def mv(src: Union[Path, str], dest: Union[Path, str]) -> None:
    """
    Move a balsam site

    balsam site mv /path/to/src /path/to/destination
    """
    cf = SiteConfig(src)

    src = Path(src).resolve()
    dest = Path(dest).resolve()

    if dest.exists():
        raise click.BadParameter(f"{dest} exists")

    shutil.move(src.as_posix(), dest.as_posix())
    client = cf.client

    site = client.Site.objects.get(id=cf.site_id)
    site.path = dest
    site.save()
    click.echo(f"Moved site to new path {dest}")


@site.command()
@click.argument("path", type=click.Path(exists=True, file_okay=False))
def rm(path: Union[str, Path]) -> None:
    """
    Remove a balsam site

    balsam site rm /path/to/site
    """
    cf = SiteConfig(path)
    client = cf.client
    site = client.Site.objects.get(id=cf.site_id)
    jobcount = client.Job.objects.filter(site_id=site.id).count()
    warning = f"This will wipe out {jobcount} jobs inside!" if jobcount else ""

    if click.confirm(f"Do you really want to destroy {Path(path).name}? {warning}"):
        site.delete()
        shutil.rmtree(path)
        click.echo(f"Deleted site {path}")


@site.command()
@click.argument("path", type=click.Path(exists=True, file_okay=False))
@click.argument("name")
def rename(path: Union[str, None], name: str) -> None:
    """
    Change the name of a balsam site
    """
    cf = SiteConfig(path)
    client = cf.client
    site = client.Site.objects.get(id=cf.site_id)
    site.name = name
    site.save()
    click.echo(f"Renamed site {site.id} to {site.name}")


@site.command()
@click.option("-v", "--verbose", is_flag=True)
def ls(verbose: bool) -> None:
    """
    List my balsam sites
    """
    client = ClientSettings.load_from_file().build_client()
    qs = client.Site.objects.all()
    if verbose:
        for site in qs:
            click.echo(str(site))
            click.echo("---\n")
    else:
        click.echo(f"{'ID':<5s}   {'Name':<26s}   {'Path':<40s}   {'Active':<4s}")
        for s in qs:
            assert s.path is not None
            assert s.last_refresh is not None
            pathstr = s.path.as_posix()
            if len(pathstr) > 37:
                pathstr = "..." + pathstr[-37:]
            active = "Yes" if is_site_active(s) else "No"
            click.echo(f"{s.id:<5d}   {s.name:<26}   {pathstr:<40}   {active:<4}")


@site.command()
def sync() -> None:
    """
    Sync changes in local settings.yml with Balsam online
    """
    cf = load_site_config()
    kill_pid = check_killable(cf, raise_exc=True)
    cf.update_site_from_config()
    click.echo("Updated site.")
    if kill_pid is not None:
        click.echo(f"Restarting Site {cf.site_path}")
        kill_site(cf, kill_pid)
        proc = start_site(cf.site_path)
        click.echo(f"Restarted Balsam site daemon [pid {proc.pid}] on {socket.gethostname()}")


@site.command()
def sample_settings() -> None:
    """
    Print a sample settings.yml site configuration
    """
    click.echo(Settings().dump_yaml())


@site.command()
def globus_login() -> None:
    """
    Get credentials for the Globus CLI

    Necessary before any Globus CLI commands which require authentication will work
    This command directs you to the page necessary to permit the Globus CLI to make API
    calls for you, and gets the OAuth2 tokens needed to use those permissions.
    """
    # if not forcing, stop if user already logged in
    if globus_auth.check_logged_in():
        click.echo("You are already logged in!")
        return

    globus_auth.do_link_auth_flow(force_new_client=True)

    click.echo("You have successfully logged in to the Globus CLI!")



import os
from typing import List, Optional

from .compute_node import ComputeNode


class SummitNode(ComputeNode):

    cpu_ids = list(range(42))
    gpu_ids = list(range(6))

    @classmethod
    def get_job_nodelist(cls) -> List["SummitNode"]:
        """
        Get all compute nodes allocated in the current job context
        """
        nodefile = os.environ["LSB_DJOB_HOSTFILE"]
        # a file containing a list of node hostnames, one per line
        # batch3
        # a01n06
        # a01n06
        # ... 1 per CPU core

        # Ignore whitespace, de-duplicate, ignore 'batch' node
        with open(nodefile) as fp:
            node_hostnames = set(line.strip() for line in fp if line.strip() and "batch" not in line)
        return [cls(host, host) for host in node_hostnames]

    @staticmethod
    def get_scheduler_id() -> Optional[int]:
        id = os.environ.get("LSB_JOBID")
        if id is not None:
            return int(id)
        return None



import os
import random
import time
from datetime import datetime
from pathlib import Path
from pprint import pprint
from typing import List

import click
import jinja2

from balsam.platform.scheduler import SlurmScheduler

os.environ["BALSAM_SITE_PATH"] = "/tmp/fakesite"

SUBMIT_PROJECT = "m3676"
SUBMIT_QUEUE = "regular"
RUN_COMMAND = "srun -n 1 --ntasks-per-node 1 --nodes 1 --cpus-per-task 32 python /global/project/projectdirs/m3676/msalim/eig/random-eig.py "
sched = SlurmScheduler()

sh_template = jinja2.Template(
    """
#!/bin/bash
#SBATCH --qos=regular
#SBATCH --reservation=xpcs-batch
#SBATCH --constraint=haswell

echo START_RUN: `date --iso-8601=ns`
cp {{ input_file }} .
echo COPIED_INPUT: `date --iso-8601=ns`
module load cray-python/3.8.5.0
export OMP_NUM_THREADS=32
{{ run_cmd }}
echo RUN_DONE: `date --iso-8601=ns`
cp *.eigs.npy {{ result_dir }}
echo COPIED_OUTPUT: `date --iso-8601=ns`
    """.strip()
)


def render_jobscript(input_file: Path, run_cmd: str, result_dir: Path) -> str:
    return sh_template.render(locals())


def submit_job(input_file: Path, workdir: Path, result_dir: Path) -> int:
    workdir.mkdir(parents=True, exist_ok=False)
    script = render_jobscript(
        input_file=input_file,
        run_cmd=RUN_COMMAND + " " + input_file.name,
        result_dir=result_dir,
    )
    with open(workdir / "job.sh", "w") as fp:
        fp.write(script)
    os.chmod(fp.name, 0o750)

    job_id = sched.submit(
        fp.name,
        project=SUBMIT_PROJECT,
        queue=SUBMIT_QUEUE,
        num_nodes=1,
        wall_time_min=5,
        cwd=workdir,
    )
    return job_id


@click.command()
@click.option("-n", "--num-nodes", required=True, type=int)
@click.option("-t", "--time-min", required=True, type=int)
@click.option("-jps", "--jobs-per-sec", required=True, type=float)
@click.option(
    "-i", "--input-file", "input_files", required=True, type=click.Path(file_okay=True, exists=True), multiple=True
)
@click.option("-e", "exp_dir", type=click.Path(exists=False), required=True)
def main(num_nodes: int, time_min: int, jobs_per_sec: float, input_files: List[Path], exp_dir: Path) -> None:
    """
    Runs batch-queue simulation
    """
    exp_dir = Path(exp_dir)
    exp_dir.mkdir(parents=True, exist_ok=False)
    result_dir = exp_dir.joinpath("results")
    result_dir.mkdir()

    input_files = [Path(inp).resolve() for inp in input_files]

    submit_delay = 1.0 / jobs_per_sec
    check_interval = 5.0
    check_steps = int(check_interval / submit_delay)
    nstep = int(time_min * 60 / submit_delay)
    pprint(locals())

    active_job_ids: List[int] = []

    for i in range(nstep):
        if len(active_job_ids) < num_nodes:
            input_file = random.choice(input_files)
            workdir = exp_dir.joinpath(input_file.with_suffix("").name).joinpath(f"{i:04d}")
            job_id = submit_job(input_file, workdir, result_dir)
            print("SUBMITTED:", workdir, datetime.now().isoformat())
            active_job_ids.append(job_id)

        if i % check_steps == 0:
            start = time.time()
            scheduler_ids = list(sched.get_statuses().keys())
            active_job_ids = [job_id for job_id in active_job_ids if job_id in scheduler_ids]
            sched_time = time.time() - start
            sleep_time = submit_delay - sched_time
        else:
            sleep_time = submit_delay
        if sleep_time > 0.0:
            time.sleep(sleep_time)


if __name__ == "__main__":
    main()



import random
from datetime import datetime, timedelta
from uuid import uuid4

import pytest

from balsam._api.app import ApplicationDefinition
from balsam.schemas import TransferItemState

GeomOpt = None
AppA = None
AppB = None
AppC = None
AppPy = None


@pytest.fixture(scope="function")
def appdef(client):
    global GeomOpt

    class GeomOpt(ApplicationDefinition):
        site = 0
        command_template = "/path/nwchem-runner {{ method }} {{ geometry }}"
        parameters = {
            "geometry": {"required": True},
            "method": {"required": False, "default": "HF"},
        }

    GeomOpt._set_client(client)
    return GeomOpt


@pytest.fixture(scope="function")
def appdef_a(client):
    global AppA

    class AppA(ApplicationDefinition):
        site = 0
        command_template = "echo hello"

    AppA._set_client(client)
    return AppA


@pytest.fixture(scope="function")
def appdef_b(client):
    global AppB

    class AppB(ApplicationDefinition):
        site = 0
        command_template = "echo hello"

    AppB._set_client(client)
    return AppB


@pytest.fixture(scope="function")
def appdef_c(client):
    global AppC

    class AppC(ApplicationDefinition):
        site = 0
        command_template = "echo hello"

    AppC._set_client(client)
    return AppC


@pytest.fixture(scope="function")
def appdef_py(client):
    global AppPy

    class AppPy(ApplicationDefinition):
        """Adds two numbers"""

        site = 0

        def run(self, x: int, y: int) -> int:
            return x + y

    AppPy._set_client(client)
    return AppPy


class TestSite:
    def test_create_and_list(self, client):
        Site = client.Site
        assert len(Site.objects.all()) == 0
        s1 = Site.objects.create(name="theta", path="/projects/foo")
        s2 = Site.objects.create(name="cooley", path="/projects/bar")
        assert len(Site.objects.all()) == 2
        assert s1.id is not None
        assert "foo" in s1.path.as_posix()
        assert "bar" in s2.path.as_posix()

    def test_create_via_save(self, client):
        Site = client.Site
        newsite = Site(name="theta", path="/projects/foo")
        assert newsite.id is None
        assert newsite._state == "creating"
        newsite.save()
        assert newsite._state == "clean"
        assert newsite.id is not None

    def test_update_status(self, client):
        Site = client.Site
        site = Site.objects.create(name="theta", path="/projects/foo")
        id = site.id
        creation_ts = site.last_refresh

        site.backfill_windows = {"default": [{"num_nodes": 31, "wall_time_min": 45}]}

        site.save()
        update_ts = site.last_refresh
        assert site.id == id
        assert update_ts > creation_ts
        assert len(site.backfill_windows) == 1

    def test_refresh_from_db(self, client):
        Site = client.Site
        handle_1 = Site.objects.create(name="theta", path="/projects/foo")
        handle_2 = Site.objects.get(id=handle_1.id)
        assert handle_2.id == handle_1.id

        handle_2.backfill_windows = {"default": [{"num_nodes": 25, "wall_time_min": 35}]}
        handle_2.save()
        assert handle_2.last_refresh > handle_1.last_refresh

        handle_1.refresh_from_db()
        assert handle_2 == handle_1

    def test_delete(self, client):
        Site = client.Site
        Site.objects.create(name="theta", path="/projects/foo")
        tempsite = Site.objects.create(name="cooley", path="/projects/bar")
        assert tempsite.id is not None
        assert len(Site.objects.all()) == 2

        tempsite.delete()
        assert tempsite.id is None
        sites = Site.objects.all()
        assert len(sites) == 1
        assert sites[0].name == "theta"

    def test_filter_on_name(self, client):
        Site = client.Site
        Site.objects.create(name="thetalogin3.alcf.anl.gov", path="/projects/foo")
        Site.objects.create(name="thetalogin4.alcf.anl.gov", path="/projects/bar")
        Site.objects.create(name="cooley", path="/projects/baz")

        cooley_only = Site.objects.filter(name="cooley")
        assert len(cooley_only) == 1

        theta_substr = Site.objects.filter(name="theta")
        assert len(theta_substr) == 0

    def test_get_by_id_returns_match(self, client):
        Site = client.Site
        site1 = Site.objects.create(name="theta", path="/projects/foo")
        retrieved = Site.objects.get(id=site1.id)
        assert site1 == retrieved

    def test_get_by_name_or_path_returns_match(self, client):
        Site = client.Site
        site1 = Site.objects.create(name="theta1", path="/projects/foo")
        site2 = Site.objects.create(name="theta2", path="/projects/bar")
        retrieved = Site.objects.get(name="theta2")
        assert retrieved == site2
        assert retrieved.id == site2.id
        retrieved = Site.objects.get(path="foo")
        assert retrieved == site1
        assert retrieved.id == site1.id

    def test_get_raises_doesnotexist(self, client):
        Site = client.Site
        with pytest.raises(Site.DoesNotExist):
            Site.objects.get(name="nonsense")

    def test_get_raises_multipleobj(self, client):
        Site = client.Site
        Site.objects.create(name="theta1", path="/projects/foo")
        Site.objects.create(name="theta2", path="/projects/foo2")
        with pytest.raises(Site.MultipleObjectsReturned):
            Site.objects.get(path="foo")

    def test_count_queryset(self, client):
        Site = client.Site
        Site.objects.create(name="theta1", path="/projects/foo")
        Site.objects.create(name="theta2", path="/projects/bar")
        Site.objects.create(name="theta3", path="/projects/baz")
        Site.objects.create(name="theta4", path="/home/bar")
        assert Site.objects.filter(path="/projects/").count() == 3
        assert Site.objects.filter(path="/home/").count() == 1


class TestApps:
    def test_create_and_list(self, client, appdef):
        App = client.App
        Site = client.Site
        assert len(App.objects.all()) == 0

        mysite = Site.objects.create(name="theta", path="/projects/foo")
        GeomOpt = appdef
        GeomOpt.site = mysite
        assert GeomOpt.__app_id__ is None
        GeomOpt.sync()
        assert len(App.objects.all()) == 1
        assert GeomOpt.__app_id__ is not None

    def test_get_by_id(self, client, appdef):
        App = client.App
        Site = client.Site
        site = Site.objects.create(name="theta", path="/projects/foo")
        GeomOpt = appdef
        GeomOpt.site = site
        GeomOpt.sync()
        assert App.objects.get(id=GeomOpt.__app_id__).name == "GeomOpt"

    def test_filter_by_site_id(self, client, appdef_a, appdef_b, appdef_c):
        App = client.App
        Site = client.Site
        site1 = Site.objects.create(name="theta", path="/projects/foo")
        site2 = Site.objects.create(name="summit", path="/projects/bar")

        A = appdef_a
        B = appdef_b
        C = appdef_c

        A.site = site1.id
        A.sync()
        B.site = site1.id
        B.sync()
        C.site = site2.id
        C.sync()

        assert App.objects.filter(site_id=site1.id).count() == 2
        assert App.objects.filter(site_id=site2.id).count() == 1

    def test_update_parameters(self, client):
        App = client.App
        Site = client.Site
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(
            site_id=site.id, name="OneApp", parameters={}, serialized_class="txt", source_code="txt"
        )
        assert app.parameters == {}
        app.parameters = {"foo": {"required": "True"}}
        app.save()

        assert App.objects.all().count() == 1
        retrieved = App.objects.get(id=app.id)
        assert "foo" in retrieved.parameters.keys()

    def test_deserializes(self, client, appdef_py):
        Site = client.Site
        site = Site.objects.create(name="theta", path="/projects/foo")
        AppPy = appdef_py
        AppPy.site = site
        AppPy.sync()

        loaded_app = client.App.objects.get(name="AppPy")
        assert "Adds two numbers" in loaded_app.description
        assert "return x + y" in loaded_app.source_code

        loaded_appdef = ApplicationDefinition.from_serialized(loaded_app)

        assert loaded_appdef.__app_id__ == AppPy.__app_id__
        assert loaded_appdef.run(None, 5, 4) == 9


class TestJobs:
    """Jobs and TransferItems"""

    def test_create(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(
            site_id=site.id,
            name="one",
            parameters={"geometry": {"required": True}},
            serialized_class="txt",
            source_code="txt",
        )

        job = Job("test/run1", app_id=app.id, parameters={"geometry": "test.xyz"}, ranks_per_node=64)
        assert job.id is None
        job.save()
        assert job.id is not None
        assert job.state == "STAGED_IN"

    def test_create_using_app_name(self, client, appdef):
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        GeomOpt = appdef
        GeomOpt.site = site
        GeomOpt.sync()

        job = Job("test/run1", app_id="GeomOpt", parameters={"geometry": "test.xyz"}, ranks_per_node=64)
        assert job.id is None
        job.save()
        assert job.id is not None
        assert job.app_id == GeomOpt.__app_id__
        assert job.state == "STAGED_IN"

    def test_create_using_app_submit_method(self, client, appdef):
        Site = client.Site
        site = Site.objects.create(name="theta", path="/projects/foo")
        GeomOpt = appdef
        GeomOpt.site = site
        GeomOpt.sync()

        job = GeomOpt.submit(workdir="test/35", geometry="h2o.xyz")
        assert job.id is not None
        assert job.get_parameters() == {"geometry": "h2o.xyz"}
        assert job.app_id == GeomOpt.__app_id__
        assert job.state == "STAGED_IN"

    def test_set_and_fetch_data(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")

        job = Job.objects.create("test/run1", app_id=app.id)
        assert job.id is not None

        job.data = {"foo": 1234}
        job.save()
        retrieved = Job.objects.get(id=job.id)
        assert retrieved.data == {"foo": 1234}

    def test_update_data(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")
        job = Job.objects.create("test/run1", app_id=app.id, data={"foo": 1234}, num_nodes=1)

        # Correct way to update a mutable field:
        data = job.data
        data["bar"] = 456
        job.data = data
        job.save()
        job.refresh_from_db()
        assert "bar" in job.data

        # Wrong way to update a mutable field:
        job.data["baz"] = 789
        job.save()
        job.refresh_from_db()
        assert "baz" not in job.data

    def test_order_limit_offset(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")
        jobs = [Job(f"test/{i}", app_id=app.id) for i in range(8)]
        random.shuffle(jobs)
        Job.objects.bulk_create(jobs)

        subset = Job.objects.all().order_by("workdir")[:4]
        assert len(subset) == 4
        assert set(job.workdir.as_posix() for job in subset) == {f"test/{i}" for i in range(4)}
        subset = Job.objects.all().order_by("workdir")[5:]
        assert len(subset) == 3
        assert set(job.workdir.as_posix() for job in subset) == {f"test/{i}" for i in range(5, 8)}
        subset = Job.objects.all().order_by("-workdir")[5:7]
        assert len(subset) == 2
        assert set(job.workdir.as_posix() for job in subset) == {"test/2", "test/1"}
        subset = Job.objects.all().order_by("workdir")[5:7]
        assert len(subset) == 2
        assert set(job.workdir.as_posix() for job in subset) == {"test/5", "test/6"}

    def test_bulk_create_and_update(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")
        jobs = [Job(f"test/{i}", app_id=app.id) for i in range(10)]

        jobs = Job.objects.bulk_create(jobs)
        assert all(job.state == "STAGED_IN" for job in jobs)

        preproc_time = datetime.utcnow()
        for job in jobs:
            job.state = "PREPROCESSED"
            job.state_data = {"message": "Skipped Preprocessing Step"}
            job.state_timestamp = preproc_time

        Job.objects.bulk_update(jobs)

        # Jobs were updated in-place:
        for job in jobs:
            assert job._state == "clean"
            assert job.state == "PREPROCESSED"
            assert job.last_update > preproc_time
            assert job.state_data is None
            assert job.state_timestamp is None

        # Jobs also updated in fresh query:
        for job in Job.objects.all():
            assert job.state == "PREPROCESSED"

    def test_children_read(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")
        parent = Job("test/parent", app_id=app.id)
        parent.save()
        child = Job("test/child", app_id=app.id, parent_ids=[parent.id])
        child.save()

        assert parent.state == "STAGED_IN"
        assert child.state == "AWAITING_PARENTS"
        assert child.parent_ids == {parent.id}
        assert child.parent_query().count() == 1

    def test_last_update_prop_changed_on_update(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(
            site_id=site.id,
            name="one",
            parameters={"geometry": {"required": False, "default": "inp.xyz"}},
            serialized_class="txt",
            source_code="txt",
        )
        job = Job("test/test", app_id=app.id, ranks_per_node=64)
        job.save()
        t1 = job.last_update

        job.num_nodes *= 2
        job.save()
        job.refresh_from_db()
        assert job.num_nodes == 2
        assert job.last_update > t1

    def test_can_view_history(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        EventLog = client.EventLog
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")
        job = Job("test/test", app_id=app.id)
        job.save()
        history = EventLog.objects.filter(job_id=job.id)
        states = [event.to_state for event in history]
        assert states == ["STAGED_IN"]

        update_time = datetime.utcnow()
        job.state = "PREPROCESSED"
        job.state_data = {"message": "Skipped Preprocess: nothing to do"}
        job.state_timestamp = update_time
        Job.objects.bulk_update([job])

        assert job.state == "PREPROCESSED"
        latest_event = EventLog.objects.filter(job_id=job.id)[0]
        assert latest_event.to_state == "PREPROCESSED"
        assert latest_event.timestamp == update_time

    def test_bulk_delete(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="One", serialized_class="txt", source_code="txt")
        jobs = [Job(f"test/{i}", app_id=app.id) for i in range(3)]
        Job.objects.bulk_create(jobs)
        assert Job.objects.count() == 3
        Job.objects.all().delete()
        assert Job.objects.count() == 0

    def test_filter_by_tags(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")
        jobs = [Job(f"test/{i}", app_id=app.id, tags={"foo": i, "bar": i * 2}) for i in range(3)]
        Job.objects.bulk_create(jobs)
        assert Job.objects.count() == 3
        qs = Job.objects.filter(tags="foo:1")
        assert qs.count() == 1
        assert qs[0].workdir.as_posix() == "test/1"
        assert Job.objects.filter(tags=["foo:2", "bar:3"]).count() == 0
        assert Job.objects.filter(tags=["foo:2", "bar:4"]).count() == 1

    def test_filter_by_id_list(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")

        foo_jobs = [Job(f"foo/{i}", app_id=app.id) for i in range(3)]
        foo_jobs = Job.objects.bulk_create(foo_jobs)
        ids = [j.id for j in foo_jobs]

        bar_jobs = [Job(f"bar/{i}", app_id=app.id) for i in range(3)]
        bar_jobs = Job.objects.bulk_create(bar_jobs)

        assert Job.objects.count() == 6
        assert Job.objects.filter(id=ids).count() == 3

    def test_state_ordering(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")
        jobs = [Job(f"foo/{i}", app_id=app.id) for i in range(4)]
        jobs = Job.objects.bulk_create(jobs)
        jobs[2].state = "PREPROCESSED"
        jobs[2].save()

        states = [job.state for job in Job.objects.all().order_by("state")]
        assert states == ["PREPROCESSED"] + ["STAGED_IN"] * 3

    def test_filter_by_workdir(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")
        jobs = [Job(f"foo/{i}", app_id=app.id) for i in range(4)]
        jobs.append(Job("bar/99", app_id=app.id))
        jobs = Job.objects.bulk_create(jobs)

        assert Job.objects.filter(workdir__contains="foo/2").count() == 1
        assert Job.objects.filter(workdir__contains="foo/8").count() == 0
        assert Job.objects.filter(workdir__contains="foo").count() == 4
        assert Job.objects.filter(workdir__contains="bar").count() == 1

    def test_filter_by_site(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site1 = Site.objects.create(name="theta1", path="/projects/foo")
        site2 = Site.objects.create(name="theta2", path="/projects/bar")
        app1 = App.objects.create(site_id=site1.id, name="one", serialized_class="txt", source_code="txt")
        app2 = App.objects.create(site_id=site2.id, name="one", serialized_class="txt", source_code="txt")

        jobs = [Job(f"foo/{i}", app_id=app1.id) for i in range(2)] + [
            Job(f"foo/{i}", app_id=app2.id) for i in range(2)
        ]
        jobs = Job.objects.bulk_create(jobs)

        # Check site filters
        assert Job.objects.filter(site_id=site1.id).count() == 2
        assert Job.objects.filter(site_id=site2.id).count() == 2
        assert Job.objects.all().count() == 4

    def test_create_by_name_and_site_path(self, client, appdef_a, appdef_b):
        App = client.App
        Site = client.Site
        Job = client.Job

        site1 = Site.objects.create(name="theta1", path="/projects/foo")
        AppA = appdef_a
        AppA.site = site1
        AppA.sync()

        site2 = Site.objects.create(name="theta2", path="/projects/bar")
        AppB = appdef_b
        AppB.site = site2
        AppB.sync()

        job1 = Job("test/1", app_id="AppA", site_name="theta1")
        job1.save()
        assert job1.id is not None
        assert job1.app_id == AppA.__app_id__

        job2 = Job("test/2", app_id="AppB", site_name="theta2")
        job2.save()
        assert job2.id is not None
        assert job2.app_id == AppB.__app_id__

        with pytest.raises(App.DoesNotExist):
            Job("test/2", app_id="AppB", site_name="theta3")

    def test_filter_by_state(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")

        jobs = [Job(f"foo/{i}", app_id=app.id) for i in range(4)]
        jobs = Job.objects.bulk_create(jobs)
        jobs[2].state = "PREPROCESSED"
        jobs[2].save()

        assert Job.objects.filter(state="STAGED_IN").count() == 3
        assert Job.objects.filter(state="PREPROCESSED").count() == 1
        assert Job.objects.filter(state__ne="STAGED_IN").count() == 1
        assert Job.objects.filter(state__ne="PREPROCESSED").count() == 3

    def test_filter_by_pending_cleanup(self, client):
        App = client.App
        Site = client.Site
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")

        jobs = [Job(f"foo/{i}", app_id=app.id) for i in range(4)]
        jobs = Job.objects.bulk_create(jobs)

        pending_cleanup = Job.objects.filter(state="STAGED_IN", pending_file_cleanup=True)
        assert pending_cleanup.count() == 4

        for job in jobs:
            job.pending_file_cleanup = False
        Job.objects.bulk_update(jobs)

        pending_cleanup = Job.objects.filter(state="STAGED_IN", pending_file_cleanup=True)
        assert pending_cleanup.count() == 0


class TestTransfers:
    def create_app_with_transfers(self, client):
        site = client.Site.objects.create(
            name="theta",
            path="/projects/foo",
            transfer_locations={"laptop": f"globus://{uuid4()}"},
        )
        app = client.App.objects.create(
            site_id=site.id,
            name="one",
            transfers={
                "input_data": {
                    "required": True,
                    "direction": "in",
                    "local_path": "inp.dat",
                },
                "output_results": {
                    "required": False,
                    "direction": "out",
                    "local_path": "results.json",
                },
            },
            serialized_class="txt",
            source_code="txt",
        )
        return app

    def test_stage_in_flow(self, client):
        Job = client.Job
        TransferItem = client.TransferItem
        app = self.create_app_with_transfers(client)
        job = Job.objects.create(
            workdir="test/test",
            app_id=app.id,
            transfers={"input_data": {"location_alias": "laptop", "path": "/path/to/input.dat"}},
        )
        assert job.state == "READY"

        transfer = TransferItem.objects.get(job_id=job.id)
        transfer.task_id = "abc"
        transfer.state = "active"
        transfer.save()

        job.refresh_from_db()
        assert job.state == "READY"

        transfer.state = "done"
        transfer.save()
        job.refresh_from_db()
        assert job.state == "STAGED_IN"

        job.state = "POSTPROCESSED"
        job.save()
        assert job.state == "JOB_FINISHED"

    def test_stage_out_flow(self, client):
        Job = client.Job
        TransferItem = client.TransferItem
        app = self.create_app_with_transfers(client)
        job = Job.objects.create(
            workdir="test/test",
            app_id=app.id,
            transfers={
                "input_data": {
                    "location_alias": "laptop",
                    "path": "/path/to/input.dat",
                },
                "output_results": {
                    "location_alias": "laptop",
                    "path": "/path/to/output.json",
                },
            },
        )
        assert job.state == "READY"
        transfers = TransferItem.objects.filter(job_id=job.id)
        assert transfers.count() == 2
        stage_in = [t for t in transfers if t.direction == "in"][0]
        stage_out = [t for t in transfers if t.direction == "out"][0]
        assert TransferItem.objects.get(state="pending") == stage_in
        assert TransferItem.objects.get(state="awaiting_job") == stage_out

        stage_in.state = "done"
        stage_in.save()
        job.refresh_from_db()
        assert job.state == "STAGED_IN"

        job.state = "POSTPROCESSED"
        job.save()
        assert job.state == "POSTPROCESSED"
        stage_out.refresh_from_db()
        assert stage_out.state == "pending"

        stage_out.state = "done"
        stage_out.save()
        job.refresh_from_db()
        assert job.state == "JOB_FINISHED"

    def test_filter_transfers_by_state(self, client):
        Job = client.Job
        TransferItem = client.TransferItem
        app = self.create_app_with_transfers(client)
        Job.objects.create(
            workdir="test/test",
            app_id=app.id,
            transfers={
                "input_data": {
                    "location_alias": "laptop",
                    "path": "/path/to/input.dat",
                },
                "output_results": {
                    "location_alias": "laptop",
                    "path": "/path/to/output.json",
                },
            },
        )
        pending = TransferItem.objects.filter(state="pending")
        assert pending.count() == 1
        pending = TransferItem.objects.filter(state=TransferItemState.pending)
        assert pending.count() == 1
        done = TransferItem.objects.filter(state=TransferItemState.done)
        assert done.count() == 0
        all_t = TransferItem.objects.filter(state=["pending", "awaiting_job"])
        assert all_t.count() == 2
        all_t = TransferItem.objects.filter(state={"pending", "awaiting_job"})
        assert all_t.count() == 2

    def test_create_transfers_shortcut(self, client):
        Job = client.Job
        TransferItem = client.TransferItem
        app = self.create_app_with_transfers(client)
        Job.objects.create(
            workdir="test/test",
            app_id=app.id,
            transfers={
                "input_data": "laptop:/path/to/input.dat",
                "output_results": "laptop:/path/to/output.json",
            },
        )
        pending = TransferItem.objects.filter(state="pending")
        assert pending.count() == 1
        assert pending[0].remote_path.as_posix() == "/path/to/input.dat"


class TestEvents:
    def setup_scenario(self, client):
        Site = client.Site
        App = client.App
        Job = client.Job
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")
        before_creation = datetime.utcnow()

        j1 = Job.objects.create(workdir="foo/1", app_id=app.id)
        j2 = Job.objects.create(workdir="foo/2", app_id=app.id)
        j3 = Job.objects.create(workdir="foo/3", app_id=app.id)

        j1.state = "PREPROCESSED"
        j1.save()

        j2.state = "PREPROCESSED"
        j2.save()
        j2.state = "RUNNING"
        j2.save()
        j2.state = "RUN_ERROR"
        j2.state_timestamp = datetime.utcnow() + timedelta(minutes=1)
        j2.save()

        j3.state = "PREPROCESSED"
        j3.save()
        j3.state = "RUNNING"
        j3.save()
        j3.state = "RUN_DONE"
        j3.state_data = {"message": "OK: done!"}
        j3.save()
        return before_creation

    def test_filter_by_job(self, client):
        self.setup_scenario(client)
        Job = client.Job
        EventLog = client.EventLog
        id = Job.objects.get(workdir__contains="foo/2").id
        qs = EventLog.objects.filter(job_id=id)
        states = [event.to_state for event in qs]
        assert states == ["RUN_ERROR", "RUNNING", "PREPROCESSED", "STAGED_IN"]

    def test_filter_by_to_state(self, client):
        self.setup_scenario(client)
        EventLog = client.EventLog
        assert EventLog.objects.filter(to_state="RUN_ERROR").count() == 1
        assert EventLog.objects.filter(to_state="STAGED_IN").count() == 3

    def test_filter_by_from_state(self, client):
        self.setup_scenario(client)
        EventLog = client.EventLog
        assert EventLog.objects.filter(from_state="CREATED").count() == 3

    def test_filter_by_to_and_from_state(self, client):
        self.setup_scenario(client)
        EventLog = client.EventLog
        qs = EventLog.objects.filter(from_state="RUNNING").filter(to_state="RUN_ERROR")
        assert qs.count() == 1
        assert qs[0] == EventLog.objects.get(from_state="RUNNING", to_state="RUN_ERROR")

    def test_filter_by_message(self, client):
        self.setup_scenario(client)
        EventLog = client.EventLog
        qs = EventLog.objects.filter(data="message:OK: done!")
        assert qs.count() == 1
        assert qs[0].to_state == "RUN_DONE"

    def test_filter_by_timestamp_range(self, client):
        self.setup_scenario(client)
        EventLog = client.EventLog
        t = datetime.utcnow() + timedelta(seconds=30)
        assert EventLog.objects.filter(timestamp_after=t).count() == 1

    def test_cannot_create_or_update(self, client):
        self.setup_scenario(client)
        EventLog = client.EventLog
        with pytest.raises(AttributeError, match="has no attribute 'create'"):
            EventLog.objects.create(
                job_id=1,
                from_state="RUNNING",
                to_state="RUN_DONE",
            )
        log = EventLog.objects.first()
        with pytest.raises(AttributeError, match="EventLog is read-only"):
            log.from_state = "CREATED"
        log.save()


class TestBatchJobs:
    def test_create(self, client):
        Site = client.Site
        BatchJob = client.BatchJob
        site = Site.objects.create(name="theta", path="/projects/foo")
        bjob = BatchJob.objects.create(
            site_id=site.id,
            project="datascience",
            queue="default",
            num_nodes=128,
            wall_time_min=30,
            job_mode="mpi",
            filter_tags={"system": "H2O", "calc_type": "energy"},
        )
        assert bjob.state == "pending_submission"
        assert bjob.id is not None

    def test_filter_by_scheduler_id(self, client):
        Site = client.Site
        BatchJob = client.BatchJob
        site = Site.objects.create(name="theta", path="/projects/foo")
        bjob = BatchJob.objects.create(
            site_id=site.id,
            project="datascience",
            queue="default",
            num_nodes=128,
            wall_time_min=30,
            job_mode="mpi",
            filter_tags={"system": "H2O", "calc_type": "energy"},
        )
        bjob.scheduler_id = 2468
        bjob.save()
        from_db = BatchJob.objects.get(site_id=site.id, scheduler_id=2468)
        assert from_db.id == bjob.id
        assert bjob.id is not None

    def test_filter_by_site_id(self, client):
        Site = client.Site
        BatchJob = client.BatchJob
        site1 = Site.objects.create(name="theta1", path="/projects/foo")
        site2 = Site.objects.create(name="theta2", path="/projects/bar")
        assert site1.id != site2.id
        BatchJob.objects.create(
            site_id=site1.id,
            project="datascience",
            queue="default",
            num_nodes=128,
            wall_time_min=30,
            job_mode="mpi",
        )
        BatchJob.objects.create(
            site_id=site2.id,
            project="datascience",
            queue="default",
            num_nodes=128,
            wall_time_min=30,
            job_mode="mpi",
        )
        from_db = BatchJob.objects.filter(site_id=site1.id)
        assert len(from_db) == 1 and from_db[0].site_id == site1.id
        from_db = BatchJob.objects.filter(site_id=site2.id)
        assert len(from_db) == 1 and from_db[0].site_id == site2.id

    def test_bulk_update(self, client):
        Site = client.Site
        BatchJob = client.BatchJob
        site = Site.objects.create(name="theta", path="/projects/foo")
        for i in range(3):
            BatchJob.objects.create(
                site_id=site.id,
                project="datascience",
                queue="default",
                num_nodes=128,
                wall_time_min=30,
                job_mode="mpi",
                filter_tags={"system": "H2O", "calc_type": "energy"},
            )

        assert BatchJob.objects.count() == 3

        bjobs = BatchJob.objects.filter(site_id=site.id)
        for job, sched_id in zip(bjobs, [123, 124, 125]):
            job.state = "queued"
            job.scheduler_id = sched_id
        BatchJob.objects.bulk_update(bjobs)

        after_update = list(BatchJob.objects.filter(site_id=site.id))
        assert after_update == list(bjobs)

    def test_delete(self, client):
        Site = client.Site
        BatchJob = client.BatchJob
        site = Site.objects.create(name="theta", path="/projects/foo")
        for i in range(3):
            BatchJob.objects.create(
                site_id=site.id,
                project="datascience",
                queue="default",
                num_nodes=128,
                wall_time_min=30,
                job_mode="mpi",
                filter_tags={"system": "H2O", "calc_type": "energy"},
            )
        assert BatchJob.objects.count() == 3

        with pytest.raises(NotImplementedError):
            BatchJob.objects.filter(site_id=site.id).delete()

        for job in BatchJob.objects.all():
            job.delete()
        assert BatchJob.objects.count() == 0

    def test_filter_by_tags(self, client):
        Site = client.Site
        BatchJob = client.BatchJob
        site = Site.objects.create(name="theta", path="/projects/foo")
        for system in ["H2O", "D2O", "NH3", "CH2O"]:
            BatchJob.objects.create(
                site_id=site.id,
                project="datascience",
                queue="default",
                num_nodes=128,
                wall_time_min=30,
                job_mode="mpi",
                filter_tags={"system": system, "calc_type": "energy"},
            )

        assert BatchJob.objects.filter(filter_tags="system:NH3").count() == 1

    def test_fetch_associated_jobs(self, client):
        Site = client.Site
        BatchJob = client.BatchJob
        App = client.App
        Job = client.Job
        Session = client.Session
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")

        batch_job = BatchJob.objects.create(
            site_id=site.id,
            project="datascience",
            queue="default",
            num_nodes=128,
            wall_time_min=30,
            job_mode="mpi",
        )
        batch_job.scheduler_id = 1234
        batch_job.state = "running"
        batch_job.save()

        for i in range(3):
            job = Job.objects.create(workdir=f"test/{i}", app_id=app.id)
            job.state = "PREPROCESSED"
            job.save()
            assert job.batch_job_id is None

        sess = Session.objects.create(batch_job_id=batch_job.id, site_id=batch_job.site_id)
        acquired = sess.acquire_jobs(
            max_wall_time_min=60,
            max_nodes_per_job=8,
            max_num_jobs=8,
            filter_tags={},
        )
        assert len(acquired) == 3

        related = Job.objects.filter(batch_job_id=batch_job.id)
        assert sorted(related, key=lambda job: job.id) == sorted(acquired, key=lambda job: job.id)


class TestSessions:
    def create_site_app(self, client):
        Site = client.Site
        App = client.App
        site = Site.objects.create(name="theta", path="/projects/foo")
        app = App.objects.create(site_id=site.id, name="one", serialized_class="txt", source_code="txt")
        return site, app

    def job(self, client, name, app, args={}, **kwargs):
        return client.Job(f"test/{name}", app_id=app.id, parameters=args, **kwargs)

    def create_jobs(self, client, app, num_jobs=3, state="PREPROCESSED"):
        jobs = [self.job(client, i, app) for i in range(num_jobs)]
        jobs = client.Job.objects.bulk_create(jobs)
        for job in jobs:
            job.state = state
        client.Job.objects.bulk_update(jobs)

    def create_sess(self, client, site):
        batch_job = client.BatchJob.objects.create(
            site_id=site.id,
            project="datascience",
            queue="default",
            num_nodes=128,
            wall_time_min=30,
            job_mode="mpi",
        )
        sess = client.Session.objects.create(batch_job_id=batch_job.id, site_id=site.id)
        return sess

    def test_create(self, client):
        before_create = datetime.utcnow()
        site = client.Site.objects.create(name="theta", path="/projects/foo")
        sess = self.create_sess(client, site)
        assert sess.heartbeat > before_create

    def test_acquire(self, client):
        site, app = self.create_site_app(client)
        self.create_jobs(client, app, num_jobs=3)
        sess = self.create_sess(client, site)

        acquired = sess.acquire_jobs(
            max_wall_time_min=60,
            max_nodes_per_job=8,
            max_num_jobs=8,
            filter_tags={},
        )
        assert len(acquired) == 3
        for job in acquired:
            assert job.batch_job_id > 0

    def test_acquire_by_app_ids(self, client):
        site, app1 = self.create_site_app(client)
        app2 = client.App.objects.create(site_id=site.id, name="two", serialized_class="txt", source_code="txt")
        self.create_jobs(client, app1, num_jobs=5)
        self.create_jobs(client, app2, num_jobs=5)
        sess = self.create_sess(client, site)

        acquired = sess.acquire_jobs(
            max_num_jobs=20,
            app_ids={app2.id},
        )
        assert len(acquired) == 5
        for job in acquired:
            assert job.app_id == app2.id

    def test_acquire_for_preprocessing(self, client):
        site, app = self.create_site_app(client)
        self.create_jobs(client, app, num_jobs=10, state="STAGED_IN")
        sess = client.Session.objects.create(site_id=site.id)
        assert sess.batch_job_id is None

        acquired = sess.acquire_jobs(
            max_wall_time_min=3600,
            max_nodes_per_job=4096,
            max_num_jobs=100,
            states=["STAGED_IN", "RUN_DONE", "RUN_ERROR", "RUN_TIMEOUT"],
        )
        assert len(acquired) == 10
        assert all(j.state == "STAGED_IN" for j in acquired)

    def test_acquire_with_filter_tags(self, client):
        site, app = self.create_site_app(client)

        job_tags = [
            dict(system="H2O", type="energy"),
            dict(system="D2O", type="energy"),
            dict(system="NO2", type="energy"),
            dict(system="H2O", type="gradient"),
        ]
        jobs = [self.job(client, i, app, tags=tags) for i, tags in enumerate(job_tags)]
        jobs = client.Job.objects.bulk_create(jobs)
        client.Job.objects.all().update(state="PREPROCESSED")

        sess = self.create_sess(client, site)
        acquired = sess.acquire_jobs(
            max_wall_time_min=60,
            max_nodes_per_job=8,
            max_num_jobs=8,
            filter_tags={"system": "H2O", "type": "energy"},
        )
        assert len(acquired) == 1
        assert acquired[0].tags == dict(system="H2O", type="energy")

    def test_tick(self, client):
        site = client.Site.objects.create(name="theta", path="/projects/foo")
        sess = self.create_sess(client, site)
        creation_time = sess.heartbeat
        sess.tick()
        sess.refresh_from_db()
        assert sess.heartbeat > creation_time

    def test_delete(self, client):
        site, app = self.create_site_app(client)
        self.create_jobs(client, app, num_jobs=3)

        sess = self.create_sess(client, site)
        acquired = sess.acquire_jobs(
            max_wall_time_min=60,
            max_nodes_per_job=8,
            max_num_jobs=8,
        )
        assert len(acquired) == 3
        assert sess.id is not None
        sess.delete()

        # After deleting session, the jobs can be re-acquired:
        assert client.Session.objects.all().count() == 0
        sess2 = self.create_sess(client, site)
        acquired = sess2.acquire_jobs(
            max_wall_time_min=60,
            max_nodes_per_job=8,
            max_num_jobs=8,
        )
        assert len(acquired) == 3



from .elastic_queue import ElasticQueueService
from .file_cleaner import FileCleanerService
from .processing import ProcessingService
from .queue_maintainer import QueueMaintainerService
from .scheduler import SchedulerService
from .transfer import TransferService

__all__ = [
    "SchedulerService",
    "ProcessingService",
    "QueueMaintainerService",
    "TransferService",
    "ElasticQueueService",
    "FileCleanerService",
]



from fastapi import status

from .util import create_app, create_site


def test_created_app_in_list_view(auth_client):
    site = create_site(auth_client)
    app = create_app(auth_client, site["id"])

    # Retrieve the app list; ensure the App shows up
    app_list = auth_client.get("/apps/")["results"]
    assert len(app_list) == 1
    assert app_list[0] == app


def test_filter_apps_by_site(auth_client):
    site1 = create_site(auth_client, name="site1", path="/site/1")
    site2 = create_site(auth_client, name="site2", path="/site/2")
    create_app(auth_client, site1["id"], name="SayHelloA")
    create_app(auth_client, site1["id"], name="SayHelloB")
    create_app(auth_client, site2["id"], name="SayHelloC")

    assert len(auth_client.get("/apps", site_id=site1["id"])["results"]) == 2
    assert len(auth_client.get("/apps", site_id=site2["id"])["results"]) == 1


def test_cannot_create_duplicate(auth_client):
    site1 = create_site(auth_client)
    site2 = create_site(auth_client, name="otherhost")
    create_app(auth_client, site_id=site1["id"], name="bar")
    create_app(auth_client, site_id=site2["id"], name="bar")
    create_app(
        auth_client,
        site_id=site1["id"],
        name="bar",
        check=status.HTTP_400_BAD_REQUEST,
    )


def test_cannot_update_duplicate(auth_client):
    site = create_site(auth_client)
    create_app(auth_client, site_id=site["id"], name="A")
    app2 = create_app(auth_client, site_id=site["id"], name="B")
    auth_client.put(f"apps/{app2['id']}", name="C")
    auth_client.put(f"apps/{app2['id']}", name="A", check=status.HTTP_400_BAD_REQUEST)
    assert auth_client.get(f"apps/{app2['id']}")["name"] == "C"


def test_delete_app(auth_client):
    site = create_site(auth_client)
    assert len(auth_client.get("/apps")["results"]) == 0
    create_app(auth_client, site_id=site["id"])
    assert len(auth_client.get("/apps")["results"]) == 1
    app2 = create_app(auth_client, site_id=site["id"], name="app")
    assert len(auth_client.get("/apps")["results"]) == 2

    auth_client.delete(f"/apps/{app2['id']}")
    assert len(auth_client.get("/apps")["results"]) == 1


def test_no_shared_app(fastapi_user_test_client):
    """client2 cannot see client1's apps by default"""
    client1, client2 = fastapi_user_test_client(), fastapi_user_test_client()
    site = create_site(client1)
    create_app(client1, site_id=site["id"])
    assert len(client1.get("/apps")["results"]) == 1
    assert len(client2.get("/apps")["results"]) == 0


def test_cannot_add_app_to_other_user_site(fastapi_user_test_client):
    client1, client2 = fastapi_user_test_client(), fastapi_user_test_client()
    site = create_site(client1)
    create_app(client2, site_id=site["id"], check=status.HTTP_404_NOT_FOUND)



import random
import time
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict

from pydantic import BaseSettings

from balsam.api import App, BatchJob, Job, Site
from balsam.schemas import BatchJobState

APP_CLASS = "eig.Eig"
RESULT_IDX = 0


class Conf(BaseSettings):
    project: str = "WorkExpFacil"
    queue: str = "R.WorkExpFacil"
    site_id: int = 11
    experiment_tag: str = "eig-backlog-theta1"

    remote_alias: str = "aps_dtn"
    result_dir: Path = Path("/gdata/lcfwork/results-2021-03-24/")
    matrix_input: Path = Path("/gdata/lcfwork/eig-inputs/5000.npy")


def submit_eig_job(
    app_id: int, workdir: Path, tags: Dict[str, str], remote_alias: str, matrix_inp_path: Path, result_dir: Path
) -> Job:
    result_path = result_dir.joinpath(matrix_inp_path.name).with_suffix(f".eig{RESULT_IDX:06d}.npy")
    transfers: Dict[str, Any] = {
        "matrix": {
            "location_alias": remote_alias,
            "path": matrix_inp_path,
        },
        "eigvals": {
            "location_alias": remote_alias,
            "path": result_path,
        },
    }
    j = Job(
        app_id=app_id,
        workdir=workdir,
        tags=tags,
        parameters={"inp_file": matrix_inp_path.name},
        transfers=transfers,
        threads_per_rank=64,
        launch_params={"cpu_affinity": "depth"},
    )
    j.save()
    return j


def run(
    site_id: int,
    app_id: int,
    experiment_tag: str,
    time_limit_min: int,
    jobs_per_sec: float,
    remote_alias: str,
    result_dir: Path,
    matrix_input: Path,
    random_kill: bool = False,
) -> None:
    global RESULT_IDX
    tags = {
        "experiment": experiment_tag,
        "job_source": remote_alias,
    }
    sleep_time = 1.0 / jobs_per_sec
    print("Will sleep for ", sleep_time, "between submissions")
    start = datetime.utcnow()
    last_kill = datetime.utcnow()

    while datetime.utcnow() - start < timedelta(minutes=time_limit_min):
        time.sleep(sleep_time)
        workdir = Path(experiment_tag) / f"{RESULT_IDX:06d}"
        job = submit_eig_job(
            app_id=app_id,
            workdir=workdir,
            tags=tags,
            remote_alias=remote_alias,
            matrix_inp_path=matrix_input,
            result_dir=result_dir,
        )
        print(f"Submitted job {job.id}")
        RESULT_IDX += 1

        if random_kill and (datetime.utcnow() - last_kill) >= timedelta(minutes=2):
            site_bjobs = list(BatchJob.objects.filter(state="running", site_id=site_id))
            if site_bjobs:
                bjob = random.choice(site_bjobs)
                bjob.state = BatchJobState.pending_deletion
                bjob.save()
                print("Killing BatchJob", bjob.id)
                last_kill = datetime.utcnow()


def main() -> None:
    conf = Conf()

    site = Site.objects.get(id=conf.site_id)
    app = App.objects.get(site_id=site.id, class_path=APP_CLASS)
    assert site.id is not None
    assert app.id is not None

    print("Starting steady run")
    run(
        site_id=site.id,
        app_id=app.id,
        experiment_tag=conf.experiment_tag,
        jobs_per_sec=1,
        time_limit_min=15,
        remote_alias=conf.remote_alias,
        result_dir=conf.result_dir,
        matrix_input=conf.matrix_input,
    )

    print("Starting too-fast run")
    run(
        site_id=site.id,
        app_id=app.id,
        experiment_tag=conf.experiment_tag,
        jobs_per_sec=3,
        time_limit_min=15,
        remote_alias=conf.remote_alias,
        result_dir=conf.result_dir,
        matrix_input=conf.matrix_input,
    )

    print("Starting random-kill  run")
    run(
        site_id=site.id,
        app_id=app.id,
        experiment_tag=conf.experiment_tag,
        jobs_per_sec=1,
        time_limit_min=15,
        random_kill=True,
        remote_alias=conf.remote_alias,
        result_dir=conf.result_dir,
        matrix_input=conf.matrix_input,
    )
    print("ALL DONE; waiting for backlog to finish normally")


if __name__ == "__main__":
    main()



from .app_run import SubprocessAppRun


class MPICHRun(SubprocessAppRun):
    """
    https://wiki.mpich.org/mpich/index.php/Using_the_Hydra_Process_Manager
    """

    def _build_cmdline(self) -> str:
        node_ids = [h for h in self._node_spec.hostnames]
        env_args = [("--env", f'{var}="{val}"') for var, val in self._envs.items()]
        nid_str = ",".join(map(str, node_ids))
        args = [
            "mpiexec",
            "-n",
            self.get_num_ranks(),
            "--ppn",
            self._ranks_per_node,
            *[arg for pair in env_args for arg in pair],
            "--hosts",
            nid_str,
            self._cmdline,
        ]
        return " ".join(str(arg) for arg in args)



from .log import Spinner, config_file_logging, config_root_logger, validate_log_level
from .process import Process
from .sighandler import SigHandler
from .time_parser import parse_to_utc

__all__ = [
    "config_file_logging",
    "config_root_logger",
    "validate_log_level",
    "Process",
    "SigHandler",
    "Spinner",
    "parse_to_utc",
]



import abc
import getpass
import subprocess
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from balsam.schemas import SchedulerBackfillWindow, SchedulerJobLog, SchedulerJobStatus

PathLike = Union[Path, str]


class SchedulerError(Exception):
    pass


class SchedulerNonZeroReturnCode(SchedulerError):
    pass


class SchedulerSubmitError(SchedulerError):
    pass


class SchedulerDeleteError(SchedulerError):
    pass


def scheduler_subproc(args: List[str], cwd: Optional[PathLike] = None) -> str:
    p = subprocess.run(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
        cwd=cwd,
    )
    if p.returncode != 0:
        raise SchedulerNonZeroReturnCode(p.stdout)
    return p.stdout


class SchedulerInterface(abc.ABC):
    def __init__(self) -> None:
        pass

    @classmethod
    @abc.abstractmethod
    def submit(
        cls,
        script_path: PathLike,
        project: str,
        queue: str,
        num_nodes: int,
        wall_time_min: int,
        cwd: Optional[PathLike] = None,
        **kwargs: Any,
    ) -> int:
        """
        Submit the script at `script_path` to a local job queue.
        Returns scheduler ID of the submitted job.
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def get_statuses(
        cls,
        project: Optional[str] = None,
        user: Optional[str] = getpass.getuser(),
        queue: Optional[str] = None,
    ) -> Dict[int, SchedulerJobStatus]:
        """
        Returns dictionary keyed on scheduler job id and a value of JobStatus for each
          job belonging to current user, project, and/or queue
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def delete_job(cls, scheduler_id: int) -> str:
        """
        Deletes the batch job matching `scheduler_id`
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def get_backfill_windows(cls) -> Dict[str, List[SchedulerBackfillWindow]]:
        """
        Returns a dictionary keyed on queue name and a value of list of
          BackfillWindow on the system for available scheduling windows
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def parse_logs(cls, scheduler_id: int, job_script_path: Optional[PathLike]) -> SchedulerJobLog:
        """
        Reads the scheduler logs to determine job metadata like start_time and end_time
        """
        raise NotImplementedError

    @classmethod
    def discover_projects(cls) -> List[str]:
        """
        Get the user's allowed/preferred allocations
        """
        project = ""
        while not project:
            project = input("Enter the name of your primary allocation: ").strip()
        return [project]


class SubprocessSchedulerInterface(SchedulerInterface, abc.ABC):
    @classmethod
    def submit(
        cls,
        script_path: PathLike,
        project: str,
        queue: str,
        num_nodes: int,
        wall_time_min: int,
        cwd: Optional[PathLike] = None,
        **kwargs: Any,
    ) -> int:
        submit_args = cls._render_submit_args(script_path, project, queue, num_nodes, wall_time_min, **kwargs)
        try:
            stdout = scheduler_subproc(submit_args, cwd=cwd)
        except SchedulerNonZeroReturnCode as exc:
            raise SchedulerSubmitError(str(exc)) from exc
        scheduler_id = cls._parse_submit_output(stdout)
        return scheduler_id

    @classmethod
    def get_statuses(
        cls,
        project: Optional[str] = None,
        user: Optional[str] = getpass.getuser(),
        queue: Optional[str] = None,
    ) -> Dict[int, SchedulerJobStatus]:
        stat_args = cls._render_status_args(project, user, queue)
        stdout = scheduler_subproc(stat_args)
        stat_dict = cls._parse_status_output(stdout)
        return stat_dict

    @classmethod
    def delete_job(cls, scheduler_id: int) -> str:
        delete_args = cls._render_delete_args(scheduler_id)
        try:
            stdout = scheduler_subproc(delete_args)
        except SchedulerNonZeroReturnCode as exc:
            raise SchedulerDeleteError(str(exc)) from exc
        return stdout

    @classmethod
    def get_backfill_windows(cls) -> Dict[str, List[SchedulerBackfillWindow]]:
        backfill_args = cls._render_backfill_args()
        stdout = scheduler_subproc(backfill_args)
        backfill_windows = cls._parse_backfill_output(stdout)
        return backfill_windows

    @classmethod
    def parse_logs(cls, scheduler_id: int, job_script_path: Optional[PathLike]) -> SchedulerJobLog:
        log_data = cls._parse_logs(scheduler_id, job_script_path)
        return log_data

    @staticmethod
    @abc.abstractmethod
    def _render_submit_args(
        script_path: PathLike, project: str, queue: str, num_nodes: int, wall_time_min: int, **kwargs: Any
    ) -> List[str]:
        pass

    @staticmethod
    @abc.abstractmethod
    def _parse_submit_output(output: str) -> int:
        pass

    @staticmethod
    @abc.abstractmethod
    def _render_status_args(project: Optional[str], user: Optional[str], queue: Optional[str]) -> List[str]:
        pass

    @staticmethod
    @abc.abstractmethod
    def _parse_status_output(output: str) -> Dict[int, SchedulerJobStatus]:
        pass

    @staticmethod
    @abc.abstractmethod
    def _render_delete_args(scheduler_id: int) -> List[str]:
        pass

    @staticmethod
    @abc.abstractmethod
    def _render_backfill_args() -> List[str]:
        pass

    @staticmethod
    @abc.abstractmethod
    def _parse_backfill_output(output: str) -> Dict[str, List[SchedulerBackfillWindow]]:
        pass

    @staticmethod
    @abc.abstractmethod
    def _parse_logs(scheduler_id: int, job_script_path: Optional[PathLike]) -> SchedulerJobLog:
        pass


__all__ = [
    "SchedulerBackfillWindow",
    "SchedulerInterface",
    "SubprocessSchedulerInterface",
    "SchedulerJobLog",
    "SchedulerJobStatus",
]



import getpass
import logging
from typing import TYPE_CHECKING, Any, Dict, Optional

from .service_base import BalsamService

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from balsam.client import RESTClient


class QueueMaintainerService(BalsamService):
    def __init__(
        self,
        client: "RESTClient",
        site_id: int,
        submit_period: int = 60,
        submit_project: str = "datascience",
        submit_queue: str = "balsam",
        job_mode: str = "mpi",
        filter_tags: Optional[Dict[str, str]] = None,
        num_queued_jobs: int = 5,
        num_nodes: int = 20,
        wall_time_min: int = 127,
    ) -> None:
        super().__init__(client=client, service_period=submit_period)
        self.site_id = site_id
        self.project = submit_project
        self.submit_queue = submit_queue
        self.job_mode = job_mode
        self.filter_tags = filter_tags
        self.num_queued_jobs = num_queued_jobs
        self.num_nodes = num_nodes
        self.wall_time_min = wall_time_min
        self.username = getpass.getuser()

    def get_next_submission(self) -> Dict[str, Any]:
        return {
            "project": self.project,
            "queue": self.submit_queue,
            "job_mode": self.job_mode,
            "filter_tags": self.filter_tags,
            "num_nodes": self.num_nodes,
            "wall_time_min": self.wall_time_min,
            "site_id": self.site_id,
        }

    def run_cycle(self) -> None:
        num_current = self.client.BatchJob.objects.filter(
            site_id=self.site_id,
            queue=self.submit_queue,
            state=["pending_submission", "queued", "running", "pending_deletion"],
        ).count()
        assert num_current is not None
        logger.debug(f"{num_current} currently active BatchJobs")
        if num_current < self.num_queued_jobs:
            sub = self.get_next_submission()
            new_job = self.client.BatchJob(**sub)
            new_job.save()
            logger.info(f"Submitted new BatchJob: {new_job}")

    def cleanup(self) -> None:
        logger.info("Exiting QueueMaintainer service")



from typing import Dict, List, cast

import click
import yaml

from balsam.schemas import BatchJobPartition, BatchJobState, JobMode

from .utils import filter_by_sites, load_client, load_site_from_selector, validate_partitions, validate_tags


@click.group()
def queue() -> None:
    """
    Submit and monitor BatchJobs (queued launcher pilots)
    """
    pass


@queue.command()
@click.option("-n", "--num-nodes", required=True, type=int)
@click.option("-t", "--wall-time-min", required=True, type=int)
@click.option("-q", "--queue", required=True)
@click.option("-A", "--project", required=True)
@click.option("-j", "--job-mode", type=click.Choice(["serial", "mpi"]), required=True)
@click.option("-tag", "--tag", "filter_tags", multiple=True, callback=validate_tags)
@click.option("-p", "--part", "partitions", multiple=True, callback=validate_partitions)
@click.option("-x", "--extra-param", "optional_params", multiple=True, callback=validate_tags)
@click.option("--site", "site_selector", default="")
def submit(
    num_nodes: int,
    wall_time_min: int,
    queue: str,
    project: str,
    job_mode: str,
    filter_tags: Dict[str, str],
    partitions: List[BatchJobPartition],
    optional_params: Dict[str, str],
    site_selector: str,
) -> None:
    """
    Submit a new BatchJob to the Site scheduler.

    1) Request 2 nodes to run all Jobs

        balsam queue submit -n 2 -t 60  -q default -A MyAllocation -j mpi

    2) Request 2 nodes to run only Jobs with certain tags

        balsam queue submit -n 2 -t 60  -q default -A alloc -j mpi -tag experiment=foo
    """
    client = load_client()
    BatchJob = client.BatchJob
    Site = client.Site

    try:
        site = load_site_from_selector(site_selector)
    except Site.DoesNotExist:
        raise click.BadParameter(f"No site matching --site {site_selector}")
    except Site.MultipleObjectsReturned:
        raise click.BadParameter(f"More than one site matching --site {site_selector}")
    assert site.id is not None

    job = BatchJob(
        site_id=site.id,
        num_nodes=num_nodes,
        wall_time_min=wall_time_min,
        queue=queue,
        project=project,
        job_mode=cast(JobMode, job_mode),
        optional_params=optional_params,
        filter_tags=filter_tags,
        partitions=partitions,
    )
    try:
        job.validate(
            site.allowed_queues,
            site.allowed_projects,
            site.optional_batch_job_params,
        )
    except ValueError as e:
        raise click.BadArgumentUsage(str(e))
    job.save()
    print(yaml.dump(job.display_dict(), sort_keys=False, indent=4))


@queue.command()
@click.option("-h", "--history", is_flag=True, default=False)
@click.option("--site", "site_selector", default="")
def ls(history: bool, site_selector: str) -> None:
    """
    List BatchJobs

    1) View current BatchJobs

        balsam queue ls

    2) View historical BatchJobs at all sites

        balsam queue ls --history --site all
    """
    client = load_client()
    BatchJob = client.BatchJob
    qs = filter_by_sites(BatchJob.objects.all(), site_selector)
    if not history:
        qs = qs.filter(state=["pending_submission", "queued", "running", "pending_deletion"])

    jobs = [j.display_dict() for j in qs]
    sites = {site.id: site for site in client.Site.objects.all()}
    for job in jobs:
        site = sites[job["site_id"]]
        path_str = site.path.as_posix()
        if len(path_str) > 27:
            path_str = "..." + path_str[-27:]
        job["site"] = f"{site.name}"

    fields = [
        "id",
        "site",
        "scheduler_id",
        "state",
        "filter_tags",
        "project",
        "queue",
        "num_nodes",
        "wall_time_min",
        "job_mode",
    ]
    rows = [[str(j[field]) for field in fields] for j in jobs]

    col_widths = [len(f) for f in fields]
    for row in rows:
        for col_idx, width in enumerate(col_widths):
            col_widths[col_idx] = max(width, len(row[col_idx]))

    for i, field in enumerate(fields):
        fields[i] = field.rjust(col_widths[i] + 1)

    print(*fields)
    for row in rows:
        for i, col in enumerate(row):
            row[i] = col.rjust(col_widths[i] + 1)
        print(*row)


@queue.command()
@click.argument("scheduler_id", type=int)
def term(scheduler_id: int) -> None:
    """
    Terminate a BatchJob by scheduler id
    """
    client = load_client()
    BatchJob = client.BatchJob

    bjob = BatchJob.objects.get(scheduler_id=scheduler_id)
    if bjob.state in [BatchJobState.queued, BatchJobState.running]:
        bjob.state = BatchJobState.pending_deletion
        bjob.save()
        click.echo("Marked BatchJob for deletion")
    else:
        click.echo("BatchJob is not queued or running")



from . import apps, batch_jobs, events, jobs, sessions, sites, transfers, users

__all__ = [
    "users",
    "sites",
    "apps",
    "jobs",
    "events",
    "sessions",
    "batch_jobs",
    "transfers",
]



import time
from datetime import datetime
from pathlib import Path
from typing import Any, Iterable, List

import pytest

from balsam._api.app import ApplicationDefinition
from balsam._api.models import BatchJob, EventLog, Job
from balsam.client import RESTClient
from balsam.config import SiteConfig
from balsam.schemas import JobMode
from balsam.schemas.batchjob import BatchJobState

from ..test_platform import get_launcher_shutdown_timeout, get_launcher_startup_timeout


def _liveness_check(batch_job: BatchJob, timeout: float = 10.0, check_period: float = 1.0) -> bool:
    """
    Poll API until Launcher is running
    """
    num_checks = int(timeout / check_period)
    for _ in range(num_checks):
        batch_job.refresh_from_db()
        if batch_job.state == "running":
            return True
        elif batch_job.state in ["pending_submission", "queued"]:
            time.sleep(check_period)
        else:
            raise RuntimeError(f"Launcher job is over: state {batch_job.state}")
    raise RuntimeError(f"Launcher did not start in {timeout} seconds.")


@pytest.fixture(scope="class", params=[JobMode.mpi, JobMode.serial])
def launcher_job(run_service: SiteConfig, request: Any) -> Iterable[BatchJob]:
    """
    Submit 1 node launcher and block until live.
    Launcher can be reused within a class.
    All dependent tests repeats for MPI mode and Serial mode.
    """
    BatchJob = run_service.client.BatchJob
    assert run_service.settings.scheduler is not None
    project = run_service.settings.scheduler.allowed_projects[0]
    queue = next(iter(run_service.settings.scheduler.allowed_queues))

    job_mode: JobMode = request.param
    batch_job = BatchJob.objects.create(
        num_nodes=1,
        wall_time_min=5,
        job_mode=job_mode,
        site_id=run_service.site_id,
        project=project,
        queue=queue,
    )
    print("Created BatchJob id:", batch_job.id, datetime.now())
    _liveness_check(batch_job, timeout=get_launcher_startup_timeout(), check_period=1.0)
    print("BatchJob started:", batch_job.id, datetime.now())
    yield batch_job
    batch_job.state = BatchJobState.pending_deletion
    batch_job.save()
    print("Killing BatchJob id:", batch_job.id, datetime.now())
    shutdown_timeout = int(get_launcher_shutdown_timeout())
    for _ in range(shutdown_timeout):
        time.sleep(1)
        batch_job.refresh_from_db()
        if batch_job.state == "finished":
            print(f"BatchJob {batch_job.id} id FINISHED at", datetime.now())
            return
    raise RuntimeError(f"Launcher did not end within {shutdown_timeout} seconds.")


@pytest.fixture(scope="function")
def live_launcher(launcher_job: BatchJob) -> bool:
    """
    Blocks until the launcher started.
    The launcher persists for the scope a class, but a quick liveness check is
    repeated before each test function.
    """
    return _liveness_check(launcher_job, timeout=1.0, check_period=1.0)


@pytest.fixture(scope="class")
def client(run_service: SiteConfig) -> RESTClient:
    return run_service.client


def poll_until_state(
    jobs: List[Job], state: str, fail_state: str = "FAILED", timeout: float = 10.0, check_period: float = 1.0
) -> bool:
    """Refresh a list of jobs until they reach the specified state"""
    num_checks = int(timeout / check_period)
    for _ in range(num_checks):
        for job in jobs:
            job.refresh_from_db()
        if all(job.state == state for job in jobs):
            return True
        if fail_state and any(job.state == fail_state for job in jobs):
            return False
        else:
            time.sleep(check_period)
    return False


@pytest.mark.usefixtures("live_launcher")
class TestSingleNode:
    @pytest.mark.parametrize("num_jobs", [3])
    def test_multi_job(self, balsam_site_config: SiteConfig, num_jobs: int, client: RESTClient) -> None:
        """
        3 hello world jobs run to completion
        """
        Hello = ApplicationDefinition.load_by_name("Hello")
        assert Hello.__app_id__ is not None
        jobs: List[Job] = []
        for i in range(num_jobs):
            job = client.Job.objects.create(
                workdir=Path(f"bar/{i}"),
                app_id=Hello,
                parameters={"name": f"world{i}!"},
                node_packing_count=2,
            )
            jobs.append(job)
        print(f"Created {num_jobs} jobs at", datetime.now())
        assert all(job.state == "STAGED_IN" for job in jobs)
        print("Start polling for jobs to reach JOB_FINISHED at", datetime.now())
        poll_until_state(jobs, "JOB_FINISHED", timeout=300.0)
        print("Polling jobs DONE at", datetime.now())
        for i, job in enumerate(jobs):
            print(f"job {i}: {job.state}")
            assert job.state in [
                "JOB_FINISHED",
                "RUN_DONE",
            ], f"Job state: {job.state}\n{list(EventLog.objects.filter(job_id=job.id))}"
            stdout = job.resolve_workdir(balsam_site_config.data_path).joinpath("job.out").read_text()
            assert f"world{i}" in stdout

    @pytest.mark.parametrize("num_jobs", [3])
    def test_py_jobs(self, balsam_site_config: SiteConfig, num_jobs: int, client: RESTClient) -> None:
        Adder = ApplicationDefinition.load_by_name("Adder")

        inputs = [(1, 1), (2, 2), (7, 3), (3, "hello")]
        jobs = [
            Adder.submit(
                workdir=Path(f"test/{i}"),
                x=inp[0],
                y=inp[1],
            )
            for i, inp in enumerate(inputs)
        ]

        print("Start polling for jobs to reach JOB_FINISHED at", datetime.now())
        poll_until_state(jobs[:3], "JOB_FINISHED", timeout=300.0)
        print("Polling jobs DONE at", datetime.now())
        poll_until_state(jobs[3:], "FAILED", timeout=10.0)

        assert jobs[0].result_nowait() == 1 + 1
        assert jobs[1].result_nowait() == 2 + 2
        assert jobs[2].result_nowait() == 7 + 3

        with pytest.raises(TypeError, match="unsupported operand type"):
            jobs[3].result_nowait()


@pytest.mark.alcf_theta
def test_cc_depth() -> None:
    assert 1



import logging
import socket
import subprocess
import sys
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union, cast

import zmq

from balsam.config import SiteConfig
from balsam.platform import TimeoutExpired
from balsam.site.launcher.node_manager import InsufficientResources, NodeManager, NodeSpec
from balsam.util import SigHandler

if TYPE_CHECKING:
    from balsam.platform.app_run import AppRun  # noqa: F401

logger = logging.getLogger(__name__)


class Worker:
    CHECK_PERIOD = 0.1
    RETRY_WINDOW = 20
    RETRY_CODES = [-11, 1, 255, 12345]
    MAX_RETRY = 3

    def __init__(
        self,
        app_run: Type["AppRun"],
        node_manager: NodeManager,
        master_host: str,
        master_port: int,
        delay_sec: int,
        error_tail_num_lines: int,
        num_prefetch_jobs: int,
        master_subproc: "Optional[subprocess.Popen[bytes]]",
    ) -> None:
        self.sig_handler = SigHandler()
        self.hostname = socket.gethostname()
        self.app_run = app_run
        self.node_manager = node_manager
        self.master_address = f"tcp://{master_host}:{master_port}"
        self.delay_sec = delay_sec
        self.error_tail_num_lines = error_tail_num_lines
        self.num_prefetch_jobs = num_prefetch_jobs
        self.master_subproc = master_subproc

        self.app_runs: Dict[int, "AppRun"] = {}
        self.start_times: Dict[int, float] = {}
        self.retry_counts: Dict[int, int] = {}
        self.job_specs: Dict[int, Dict[str, Any]] = {}
        self.node_specs: Dict[int, NodeSpec] = {}
        self.runnable_cache: Dict[int, Dict[str, Any]] = {}

    def cleanup_proc(self, id: int, timeout: float = 0) -> None:
        self.kill(id, timeout=timeout)
        self.node_manager.free(id)
        del self.app_runs[id]
        del self.start_times[id]
        del self.retry_counts[id]
        del self.job_specs[id]
        del self.node_specs[id]

    def check_retcodes(self) -> List[Tuple[int, Optional[int]]]:
        id_retcodes = []
        for id, proc in self.app_runs.items():
            retcode = proc.poll()
            id_retcodes.append((id, retcode))
        return id_retcodes

    def log_error_tail(self, id: int, retcode: int) -> str:
        tail = self.app_runs[id].tail_output(self.error_tail_num_lines)
        logmsg = f"Job {id} nonzero return {retcode}:\n {tail}"
        logger.error(logmsg)
        return tail

    def can_retry(self, id: int, retcode: int) -> bool:
        if retcode in self.RETRY_CODES:
            elapsed = time.time() - self.start_times[id]
            retry_count = self.retry_counts[id]
            if elapsed < self.RETRY_WINDOW and retry_count <= self.MAX_RETRY:
                logmsg = (
                    f"Job {id} can retry task (err occured after {elapsed:.2f} sec; "
                    f"attempt {self.retry_counts[id]}/{self.MAX_RETRY})"
                )
                logger.error(logmsg)
                return True
        return False

    def kill(self, id: int, timeout: float = 0) -> None:
        p = self.app_runs[id]
        if p.poll() is None:
            p.terminate()
            logger.debug(f"worker {self.hostname} sent TERM to {id}...waiting on shutdown")
            try:
                p.wait(timeout=timeout)
            except TimeoutExpired:
                p.kill()

    def launch_run(self, id: int) -> None:
        job_spec = self.job_specs[id].copy()
        node_spec = self.node_specs[id]
        job_spec.pop("id")
        job_spec.pop("node_occupancy")

        logger.debug(f"Job {id} WORKER_START")
        proc = self.app_run(
            **job_spec,
            node_spec=node_spec,
            ranks_per_node=1,
            launch_params={},
            outfile_path=Path(job_spec["cwd"]).joinpath("job.out"),
        )
        proc.start()
        self.app_runs[id] = proc

    def handle_error(self, id: int, retcode: int) -> Union[Tuple[int, str], str]:
        tail = self.log_error_tail(id, retcode)

        if not self.can_retry(id, retcode):
            self.cleanup_proc(id)
            return (retcode, tail)
        else:
            self.start_times[id] = time.time()
            self.retry_counts[id] += 1
            self.launch_run(id)
            return "running"

    def poll_processes(self) -> Tuple[List[int], List[Tuple[int, int, str]]]:
        done, error = [], []
        for id, retcode in self.check_retcodes():
            if retcode is None:
                continue
            elif retcode == 0:
                done.append(id)
                logger.debug(f"Job {id} WORKER_DONE")
                self.cleanup_proc(id)
            else:
                logger.info(f"Job {id} WORKER_ERROR")
                status = self.handle_error(id, retcode)
                if status != "running":
                    retcode, tail = cast(Tuple[int, str], status)
                    error.append((id, retcode, tail))
        return done, error

    def exit(self) -> None:
        ids = list(self.app_runs.keys())
        for id in ids:
            self.cleanup_proc(id, timeout=self.CHECK_PERIOD)
        self.socket.setsockopt(zmq.LINGER, 0)
        self.socket.close(linger=0)
        self.context.term()  # type: ignore

    def start_jobs(self) -> List[int]:
        started_ids = []

        for id, job_spec in self.runnable_cache.items():
            try:
                node_spec = self.node_manager.assign_from_params(
                    **job_spec,
                    num_nodes=1,
                    ranks_per_node=1,
                )
            except InsufficientResources:
                continue
            else:
                self.job_specs[id] = job_spec
                self.node_specs[id] = node_spec
                self.start_times[id] = time.time()
                self.retry_counts[id] = 1
                self.launch_run(id)
                started_ids.append(id)

        self.runnable_cache = {k: v for k, v in self.runnable_cache.items() if k not in started_ids}
        return started_ids

    def cycle(self) -> bool:
        """Run a cycle of Job dispatch. Returns True if worker should continue; False if time to exit."""
        done_ids, errors = self.poll_processes()
        started_ids = self.start_jobs()
        request_num_jobs = max(0, self.num_prefetch_jobs - len(self.runnable_cache))

        msg = {
            "source": self.hostname,
            "started": started_ids,
            "done": done_ids,
            "error": errors,
            "request_num_jobs": request_num_jobs,
        }
        self.socket.send_json(msg)
        logger.debug("Worker awaiting response...")
        response_msg = self.socket.recv_json()
        logger.debug("Worker response received")

        if response_msg.get("exit"):
            logger.info(f"Worker {self.hostname} received exit message: break")
            return False

        if response_msg.get("new_jobs"):
            self.runnable_cache.update({job["id"]: job for job in response_msg["new_jobs"]})

        logger.debug(
            f"{self.hostname} fraction available: {self.node_manager.aggregate_free_nodes()} "
            f"[{len(self.runnable_cache)} additional prefetched "
            f"jobs in cache]"
        )
        return True

    def run(self) -> None:
        self.context = zmq.Context()  # type: ignore
        self.context.setsockopt(zmq.LINGER, 0)  # type: ignore
        self.socket = self.context.socket(zmq.REQ)  # type: ignore
        self.socket.connect(self.master_address)
        logger.debug(f"Worker connected to {self.master_address}")

        # Run the Worker loop until master sends "exit" message. Does not quit on SIGTERM.
        while self.cycle():
            # If SIGTERM has been received, pass onto master and keep waiting for "exit"
            time.sleep(self.delay_sec)
            if self.sig_handler.is_set() and self.master_subproc is not None:
                self.master_subproc.terminate()
                logger.info("Signal: forwarded SIGTERM to master subprocess.")


def launch_master_subprocess() -> "subprocess.Popen[bytes]":
    args = [sys.executable] + sys.argv + ["--run-master"]
    return subprocess.Popen(args)


def worker_main(
    master_host: str,
    master_port: int,
    log_filename: str,
) -> None:
    site_config = SiteConfig()
    hostname = socket.gethostname()
    master_proc = None

    SigHandler()
    site_config.enable_logging("serial_mode", filename=log_filename + f".{hostname}")
    if hostname == master_host:
        logger.info(f"Launching master subprocess on {hostname}")
        master_proc = launch_master_subprocess()
    else:
        logger.info(f"Worker on {hostname} will connect to remote master on {master_host}")

    launch_settings = site_config.settings.launcher
    node_cls = launch_settings.compute_node
    nodes = [node for node in node_cls.get_job_nodelist() if node.hostname == hostname]
    node_manager = NodeManager(nodes, allow_node_packing=True)
    worker = Worker(
        app_run=launch_settings.local_app_launcher,
        node_manager=node_manager,
        master_host=master_host,
        master_port=master_port,
        delay_sec=launch_settings.delay_sec,
        error_tail_num_lines=launch_settings.error_tail_num_lines,
        num_prefetch_jobs=launch_settings.serial_mode_prefetch_per_rank,
        master_subproc=master_proc,
    )

    try:
        logger.debug("Launching worker")
        # Worker does not quit on SIGTERM; wait until master has sent "exit"
        worker.run()
    except:  # noqa
        raise
    finally:
        worker.exit()
        if master_proc is not None:
            try:
                master_proc.wait(timeout=5)
                logger.info("master process shutdown OK")
            except subprocess.TimeoutExpired:
                logger.warning("Force-killing master process")
                master_proc.kill()



from typing import Tuple

from sqlalchemy.orm import Query, Session

from balsam import schemas
from balsam.server import models
from balsam.server.routers.filters import EventLogQuery
from balsam.server.utils import Paginator


def fetch(
    db: Session, owner: schemas.UserOut, paginator: Paginator[models.LogEvent], filterset: EventLogQuery
) -> "Tuple[int, Query[models.LogEvent]]":
    qs = db.query(models.LogEvent).join(models.Job).join(models.App).join(models.Site)  # type: ignore
    qs = qs.filter(models.Site.owner_id == owner.id)
    qs = filterset.apply_filters(qs)
    count = qs.group_by(models.LogEvent.id).count()
    events = paginator.paginate(qs)
    return count, events



import logging
from math import ceil
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Optional, Tuple, Type, TypeVar

from .model import BalsamModel
from .query import Query

if TYPE_CHECKING:
    from balsam.client import RESTClient


logger = logging.getLogger(__name__)
T = TypeVar("T", bound=BalsamModel)


class Manager(Generic[T]):
    _model_class: Type[T]
    _query_class: Type[Query[T]]
    _bulk_create_enabled: bool
    _bulk_update_enabled: bool
    _bulk_delete_enabled: bool
    _paginated_list_response: bool
    _api_path: str

    def __init__(self, client: "RESTClient") -> None:
        self._client = client
        self._model_class.objects = self

    def all(self) -> Query[T]:
        return self._query_class(manager=self)

    def count(self) -> Optional[int]:
        return self.all().count()

    def first(self) -> T:
        return self.all().first()

    def _create(self, instance: Optional[T] = None, **data: Any) -> T:
        # Pass raw dict through the BalsamModel constructor for validation
        if instance is None:
            instance = self._model_class(**data)
        if self._bulk_create_enabled:
            created_list = self.bulk_create([instance])
            created = created_list[0]
        else:
            assert instance._create_model is not None
            data = instance._create_model.dict()
            created = self._client.post(self._api_path, **data)
            created = self._model_class._from_api(created)
        return created

    def bulk_create(self, instances: List[T]) -> List[T]:
        """Returns a list of newly created instances"""
        if not self._bulk_create_enabled:
            raise NotImplementedError("The {self._model_class.__name__} API does not offer bulk_create")

        if not isinstance(instances, list):
            raise TypeError(f"instances must be a list of {self._model_class.__name__} instances")

        assert self._model_class._create_model_cls is not None
        assert all(
            isinstance(obj._create_model, self._model_class._create_model_cls) for obj in instances
        ), f"bulk_create requires all items to be instances of {self._model_class.__name__}"

        data_list = [obj._create_model.dict() for obj in instances]  # type: ignore # (checked in the assert above)
        response_data = self._client.bulk_post(self._api_path, data_list)
        # Cannot update in-place: no IDs to perform the mapping yet
        return [self._model_class._from_api(dat) for dat in response_data]

    def bulk_update(self, instances: List[T]) -> None:
        """
        Perform a bulk patch of instances from the modified `instances` list and set of
        `update_fields`. Modifies the instances list in-place and returns None.
        """
        # TODO: validate update_fields
        if not self._bulk_update_enabled:
            raise NotImplementedError("The {self._model_class.__name__} API does not offer bulk_update")

        patch_list = []
        for obj in instances:
            if obj._update_model is not None:
                patch = {"id": obj.id, **obj._update_model.dict(exclude_unset=True)}
                patch_list.append(patch)

        if not patch_list:
            logger.debug("bulk-update: patch_list is empty (nothing to update!)")
            return

        logger.debug(f"Performing bulk-update to {self._api_path}\n{patch_list}")
        response_data = self._client.bulk_patch(self._api_path, patch_list)
        response_map = {item["id"]: item for item in response_data}
        logger.debug(f"bulk-update response map: {response_map}")
        # Use response_map to update instances in-place
        for obj in instances:
            if obj.id in response_map:
                logger.debug(f"refreshing object id={obj.id} from response_map")
                obj._refresh_from_dict(response_map[obj.id])

    def _build_query_params(
        self,
        filters: Dict[str, Any],
        ordering: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Dict[str, Any]:
        d = {}
        d.update(filters)
        if ordering is not None:
            d.update(ordering=ordering)
        if limit is not None:
            d.update(limit=limit)
        if offset is not None:
            d.update(offset=offset)
        return d

    @staticmethod
    def _chunk_filters(filters: Dict[str, Any]) -> List[Dict[str, Any]]:
        chunk_size = 512
        chunked_filter = next(
            (
                (name, filter)
                for (name, filter) in filters.items()
                if isinstance(filter, (list, tuple, set)) and len(filter) > chunk_size
            ),
            None,
        )
        if chunked_filter is None:
            filter_chunks = [filters]
        else:
            name, filter = chunked_filter
            filter = list(filter)
            nchunk = ceil(len(filter) / chunk_size)
            chunks = [filter[i * chunk_size : (i + 1) * chunk_size] for i in range(nchunk)]
            filter_chunks = [{**filters, name: chunk} for chunk in chunks]
        return filter_chunks

    def _unpack_list_response(self, response_data: Dict[str, Any]) -> Tuple[Optional[int], List[Dict[str, Any]]]:
        if self._paginated_list_response:
            count = response_data["count"]
            results = response_data["results"]
        else:
            count = None
            results = response_data
        return count, results

    def _get_list(
        self,
        filters: Dict[str, Any],
        ordering: Optional[str],
        limit: Optional[int],
        offset: Optional[int],
    ) -> Tuple[List[T], Optional[int]]:

        filter_chunks = self._chunk_filters(filters)
        full_count: Optional[int] = 0
        full_results: List[Dict[str, Any]] = []

        # Added complexity: we handle the case that one URL query
        # parameter is too long: chunk query into multiple GETs passing subsets
        # of the sequence (e.g. filter by list of 100k job ids will result in 196 requests
        # being stitched together)
        for filter_chunk in filter_chunks:
            query_params = self._build_query_params(filter_chunk, ordering, limit, offset)
            response_data = self._client.get(self._api_path, **query_params)
            count, results = self._unpack_list_response(response_data)
            if count is not None and full_count is not None:
                full_count += count
            else:
                full_count = None
            full_results.extend(results)
        if ordering and len(filter_chunks) > 1:
            order_key, reverse = (ordering.lstrip("-"), True) if ordering.startswith("-") else (ordering, False)
            full_results = sorted(full_results, key=lambda r: r[order_key], reverse=reverse)  # type: ignore
        instances = [self._model_class._from_api(dat) for dat in full_results]
        return instances, full_count

    def _do_update(self, instance: T) -> None:
        assert instance._update_model is not None
        update_data = instance._update_model.dict(exclude_unset=True)
        response_data = self._client.put(
            self._api_path + f"{instance.id}",
            **update_data,
        )
        instance._refresh_from_dict(response_data)

    def _do_bulk_update_query(self, patch: Dict[str, Any], filters: Dict[str, Any]) -> List[T]:
        if not self._bulk_update_enabled:
            raise NotImplementedError(f"The {self._model_class.__name__} API does not offer bulk updates")
        query_params = self._build_query_params(filters)
        response_data = self._client.bulk_put(self._api_path, patch, **query_params)
        instances = [self._model_class._from_api(dat) for dat in response_data]
        return instances

    def _do_delete(self, instance: T) -> None:
        self._client.delete(self._api_path + f"{instance.id}")
        if instance._read_model is not None:
            instance._read_model.id = None  # type: ignore

    def _do_bulk_delete(self, filters: Dict[str, Any]) -> None:
        if not self._bulk_delete_enabled:
            raise NotImplementedError(f"The {self._model_class.__name__} API does not offer bulk deletes")
        query_params = self._build_query_params(filters)
        self._client.bulk_delete(self._api_path, **query_params)



import json
import logging
import time
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Type, Union

import click

from balsam._api.app import ApplicationDefinition
from balsam.config import SiteConfig
from balsam.platform import TimeoutExpired
from balsam.schemas import DeserializeError, JobState
from balsam.site import BulkStatusUpdater, SynchronousJobSource
from balsam.site.launcher.node_manager import NodeManager
from balsam.site.launcher.util import countdown_timer_min
from balsam.util import SigHandler

logger = logging.getLogger("balsam.site.launcher.mpi_mode")

if TYPE_CHECKING:
    from balsam._api.models import Job
    from balsam.platform.app_run import AppRun


class Launcher:
    def __init__(
        self,
        data_dir: Path,
        idle_ttl_sec: int,
        app_run: Type["AppRun"],
        node_manager: NodeManager,
        job_source: SynchronousJobSource,
        status_updater: BulkStatusUpdater,
        wall_time_min: int,
        delay_sec: int,
        error_tail_num_lines: int,
        max_concurrent_runs: int,
    ) -> None:
        self.data_dir = data_dir
        self.idle_ttl_sec = idle_ttl_sec
        self.error_tail_num_lines = error_tail_num_lines
        self.app_run = app_run
        self.node_manager = node_manager
        self.job_source = job_source
        self.status_updater = status_updater
        self.timer = countdown_timer_min(max(1, wall_time_min - 2), delay_sec)

        self.status_updater.start()
        self.job_source.start()
        self.active_runs: Dict[int, "AppRun"] = {}
        self.idle_time: Optional[float] = None
        self.max_concurrent_runs = max_concurrent_runs

        self.sig_handler = SigHandler()

    def time_step(self) -> None:
        try:
            min_left = next(self.timer)
        except StopIteration:
            self.sig_handler.set()
            return

        m, s = map(int, divmod(min_left * 60, 60))
        logger.debug(f"{m:02d}m:{s:02d}s remaining")

    def check_exit(self) -> None:
        if not self.active_runs:
            if self.idle_time is None:
                self.idle_time = time.time()
            elif time.time() - self.idle_time > self.idle_ttl_sec:
                self.sig_handler.set()
                logger.info(f"Exceeded {self.idle_ttl_sec} sec TTL: shutting down because nothing to do")
        else:
            self.idle_time = None

    def create_run(self, job: "Job") -> "AppRun":
        app_cls = ApplicationDefinition.load_by_id(job.app_id)
        app = app_cls(job)
        workdir = self.data_dir.joinpath(app.job.workdir)

        preamble = app.shell_preamble()
        app_command = app.get_arg_str()
        environ_vars = app.get_environ_vars()

        # assign workers
        node_spec = self.node_manager.assign(job)
        # start run
        run = self.app_run(
            cmdline=app_command,
            preamble=preamble,
            envs=environ_vars,
            cwd=workdir,
            outfile_path=workdir.joinpath("job.out"),
            node_spec=node_spec,
            ranks_per_node=job.ranks_per_node,
            threads_per_rank=job.threads_per_rank,
            threads_per_core=job.threads_per_core,
            launch_params=job.launch_params,
            gpus_per_rank=int(job.gpus_per_rank),
        )
        return run

    def acquire_jobs(self) -> List["Job"]:
        max_nodes_per_job = self.node_manager.count_empty_nodes()
        max_aggregate_nodes = self.node_manager.aggregate_free_nodes()
        max_num_to_acquire = max(0, self.max_concurrent_runs - len(self.active_runs))
        if not self.node_manager.allow_node_packing:
            max_num_to_acquire = min(max_num_to_acquire, int(max_aggregate_nodes))
        if max_aggregate_nodes < 0.01:
            return []

        acquired = self.job_source.get_jobs(
            max_num_jobs=max_num_to_acquire,
            max_nodes_per_job=max_nodes_per_job,
            max_aggregate_nodes=max_aggregate_nodes,
        )
        if acquired:
            logger.info(
                f"Job Acqusition: {max_nodes_per_job} empty nodes; {max_aggregate_nodes} aggregate free nodes; "
                f"requested up to {max_num_to_acquire} jobs [node packing allowed: {self.node_manager.allow_node_packing}]; "
                f"Acquired {len(acquired)} jobs."
            )
        return acquired

    def launch_runs(self) -> None:
        acquired = self.acquire_jobs()
        for job in acquired:
            assert job.id is not None
            try:
                run = self.create_run(job)
            except DeserializeError as exc:
                logger.exception(f"Failed to deserialize for Job(id={job.id}, workdir={job.workdir}): {exc}")
                self.status_updater.put(
                    job.id,
                    state=JobState.failed,
                    state_timestamp=datetime.utcnow(),
                    state_data={
                        "message": "An exception occured while loading the app or parameters",
                        "exception": str(exc),
                    },
                )
            else:
                run.start()
                self.status_updater.put(
                    job.id,
                    state=JobState.running,
                    state_timestamp=datetime.utcnow(),
                    state_data={"num_nodes": float(job.num_nodes) / job.node_packing_count},
                )
                self.active_runs[job.id] = run

    def check_run(self, run: "AppRun") -> Dict[str, Any]:
        retcode = run.poll()
        if retcode is None:
            return {"state": "RUNNING"}
        elif retcode == 0:
            return {"state": "RUN_DONE", "state_timestamp": datetime.utcnow()}
        else:
            tail = run.tail_output(nlines=self.error_tail_num_lines)
            logger.info(f"Run error: {tail}")
            return {
                "state": "RUN_ERROR",
                "state_timestamp": datetime.utcnow(),
                "state_data": {"returncode": retcode, "error": tail},
            }

    @staticmethod
    def timeout_kill(runs: Iterable["AppRun"], timeout: float = 10) -> None:
        start = time.time()
        for run in runs:
            remaining = max(0, timeout - (time.time() - start))
            try:
                run.wait(timeout=remaining)
            except TimeoutExpired:
                run.kill()

    def update_states(self, timeout: bool = False) -> None:
        remaining_runs = {}
        for id, run in self.active_runs.items():
            status = self.check_run(run)
            if status["state"] == "RUNNING" and timeout:
                run.terminate()
                self.status_updater.put(id, state=JobState.run_timeout, state_timestamp=datetime.utcnow())
                self.node_manager.free(id)
                remaining_runs[id] = run
            elif status["state"] == "RUNNING":
                remaining_runs[id] = run
            else:
                self.status_updater.put(id, **status)
                self.node_manager.free(id)
        if timeout:
            self.timeout_kill(remaining_runs.values())
            self.active_runs = {}
        else:
            self.active_runs = remaining_runs

    def run(self) -> None:
        try:
            while not self.sig_handler.is_set():
                self.time_step()
                self.launch_runs()
                self.update_states()
                self.check_exit()
        except:  # noqa
            raise
        finally:
            logger.info("Launcher starting shutdown sequence")
            self.job_source.terminate()
            logger.info("Timing out active runs")
            self.update_states(timeout=True)
            self.status_updater.terminate()
            self.job_source.join()
            self.status_updater.join()

    def timeout_runs(self) -> None:
        for run in self.active_runs.values():
            run.terminate()


@click.command()
@click.option("--wall-time-min", type=int)
@click.option("--log-filename")
@click.option("--node-ids")
@click.option("--filter-tags")
def main(
    wall_time_min: int,
    log_filename: str,
    node_ids: str,
    filter_tags: str,
) -> None:
    site_config = SiteConfig()
    site_config.enable_logging("mpi_mode", filename=log_filename)
    filter_tags_dict: Dict[str, str] = json.loads(filter_tags)
    node_ids_list: List[Union[int, str]] = json.loads(node_ids)
    if filter_tags_dict:
        logger.info(f"Launcher filtering for tags: {filter_tags_dict}")

    launch_settings = site_config.settings.launcher
    node_cls = launch_settings.compute_node
    nodes = [node for node in node_cls.get_job_nodelist() if node.node_id in node_ids_list]
    node_manager = NodeManager(nodes, allow_node_packing=launch_settings.mpirun_allows_node_packing)

    ApplicationDefinition._set_client(site_config.client)
    try:
        ApplicationDefinition.load_by_site(site_config.site_id)  # Warms the cache
    except DeserializeError as exc:
        logger.warning(
            f"At least one App registered at this Site failed to deserialize: {exc}. "
            "Jobs running with this App will FAIL.  Please fix the App classes and/or the "
            "Balsam Site environment, and double check that the Apps can load OK."
        )

    scheduler_id = node_cls.get_scheduler_id()
    job_source = SynchronousJobSource(
        client=site_config.client,
        site_id=site_config.site_id,
        filter_tags=filter_tags_dict,
        max_wall_time_min=wall_time_min,
        scheduler_id=scheduler_id,
    )
    status_updater = BulkStatusUpdater(site_config.client)

    launcher = Launcher(
        data_dir=site_config.data_path,
        idle_ttl_sec=launch_settings.idle_ttl_sec,
        delay_sec=launch_settings.delay_sec,
        app_run=launch_settings.mpi_app_launcher,
        node_manager=node_manager,
        job_source=job_source,
        status_updater=status_updater,
        wall_time_min=wall_time_min,
        error_tail_num_lines=launch_settings.error_tail_num_lines,
        max_concurrent_runs=launch_settings.max_concurrent_mpiruns,
    )
    launcher.run()
